"""Parameters estimation."""

from __future__ import annotations

from typing import Any, List, Union
import numpy as np
import scipy as sp
import scipy.special as _spsp
from numpy import ndarray


ninf = 1e-5
MAXIT = 20
EPS = 1e-6
# Euler's constant
EU = 0.577215664901532861


class Lmoments:
    """Lmoments.

        Hosking (1990) introduced the concept of L-moments, which are quantities that
        can be directly interpreted as scale and shape descriptors of probability distributions
        The L-moments of order r, denoted by λr.

    λ1 = α0 = β0
    λ2 = α0 - 2α1 = 2β1 - β0
    λ3 = α0 - 6α1 + 6α2 = 6β2 - 6β1 + β0
    λ4 = α0 - 12α1 + 30α2 - 20α3 = 20β3 - 30β2 + 12β1 - β0
    """

    def __init__(self, data):
        self.data = data
        pass

    def Lmom(self, nmom=5):
        """Calculates the Lmoments."""
        if nmom <= 5:
            var = self._samlmusmall(nmom)
        else:
            var = self._samlmularge(nmom)

        return var

    @staticmethod
    def _comb(N, k):
        """sum [(N-j)/(j+1)]"""
        if (k > N) or (N < 0) or (k < 0):
            val = 0
        else:
            val = 1
            for j in range(min(k, N - k)):
                val = (val * (N - j)) // (j + 1)  # // is floor division
        return val

    def _samlmularge(self, nmom: int = 5) -> list[ndarray | float | int | Any]:
        """Large sample L moment."""

        x = self.data
        if nmom <= 0:
            raise ValueError("Invalid number of Sample L-Moments")

        x = sorted(x)
        n = len(x)

        if n < nmom:
            raise ValueError("Insufficient length of data for specified nmoments")

        # Calculate first order
        coef_l1 = 1.0 / self._comb(n, 1)
        sum_l1 = sum(x)
        lmoments = [coef_l1 * sum_l1]

        if nmom == 1:
            return lmoments[0]

        # Setup comb table, where comb[i][x] refers to comb(x,i)
        comb = []
        for i in range(1, nmom):
            comb.append([])
            for j in range(n):
                comb[-1].append(self._comb(j, i))

        for mom in range(2, nmom + 1):
            coefl = 1.0 / mom * 1.0 / self._comb(n, mom)
            xtrans = []
            for i in range(0, n):
                coef_temp = []
                for _ in range(0, mom):
                    coef_temp.append(1)

                for j in range(0, mom - 1):
                    coef_temp[j] = coef_temp[j] * comb[mom - j - 2][i]

                for j in range(1, mom):
                    coef_temp[j] = coef_temp[j] * comb[j - 1][n - i - 1]

                for j in range(0, mom):
                    coef_temp[j] = coef_temp[j] * self._comb(mom - 1, j)

                for j in range(0, int(0.5 * mom)):
                    coef_temp[j * 2 + 1] = -coef_temp[j * 2 + 1]
                coef_temp = sum(coef_temp)
                xtrans.append(x[i] * coef_temp)

            if mom > 2:
                lmoments.append(coefl * sum(xtrans) / lmoments[1])
            else:
                lmoments.append(coefl * sum(xtrans))
        return lmoments

    def _samlmusmall(self, nmom: int = 5) -> list[ndarray | float | int | Any]:
        """Small sample L-Moments."""
        sample = self.data

        if nmom <= 0:
            raise ValueError("Invalid number of Sample L-Moments")

        sample = sorted(sample)
        n = len(sample)

        if n < nmom:
            raise ValueError("Insufficient length of data for specified nmoments")

        # coefl1 = 1.0 / self._comb(n, 1)  # coefl1 = 1/n
        # suml1 = sum(sample)
        # l_moment_1 = coefl1 * suml1  # l_moment_1 = mean(sample)
        l_moment_1 = np.mean(sample)
        if nmom == 1:
            return [l_moment_1]

        # comb terms appear elsewhere, this will decrease calc time
        # for nmom > 2, and shouldn't decrease time for nmom == 2
        # comb(sample,1) = sample
        # for i in range(1,n+1):
        # #        comb1.append(comb(i-1,1))
        # #        comb2.append(comb(n-i,1))
        # Can be simplified to comb1 = range(0,n)

        comb1 = range(0, n)
        comb2 = range(n - 1, -1, -1)

        coefl2 = 0.5 * 1.0 / self._comb(n, 2)
        xtrans = []
        for i in range(0, n):
            coef_temp = comb1[i] - comb2[i]
            xtrans.append(coef_temp * sample[i])

        l_moment_2 = coefl2 * sum(xtrans)

        if nmom == 2:
            return [l_moment_1, l_moment_2]

        # Calculate Third order
        # comb terms appear elsewhere, this will decrease calc time
        # for nmom > 2, and shouldn't decrease time for nmom == 2
        # comb3 = comb(i-1,2)
        # comb4 = comb3.reverse()
        comb3 = []
        comb4 = []
        for i in range(0, n):
            comb_temp = self._comb(i, 2)
            comb3.append(comb_temp)
            comb4.insert(0, comb_temp)

        coefl3 = 1.0 / 3 * 1.0 / self._comb(n, 3)
        xtrans = []
        for i in range(0, n):
            coef_temp = comb3[i] - 2 * comb1[i] * comb2[i] + comb4[i]
            xtrans.append(coef_temp * sample[i])

        l_moment_3 = coefl3 * sum(xtrans) / l_moment_2

        if nmom == 3:
            return [l_moment_1, l_moment_2, l_moment_3]

        # Calculate Fourth order
        # comb5 = comb(i-1,3)
        # comb6 = comb(n-i,3)
        comb5 = []
        comb6 = []
        for i in range(0, n):
            comb_temp = self._comb(i, 3)
            comb5.append(comb_temp)
            comb6.insert(0, comb_temp)

        coefl4 = 1.0 / 4 * 1.0 / self._comb(n, 4)
        xtrans = []
        for i in range(0, n):
            coef_temp = (
                comb5[i] - 3 * comb3[i] * comb2[i] + 3 * comb1[i] * comb4[i] - comb6[i]
            )
            xtrans.append(coef_temp * sample[i])

        l_moment_4 = coefl4 * sum(xtrans) / l_moment_2

        if nmom == 4:
            return [l_moment_1, l_moment_2, l_moment_3, l_moment_4]

        # Calculate Fifth order
        comb7 = []
        comb8 = []
        for i in range(0, n):
            comb_temp = self._comb(i, 4)
            comb7.append(comb_temp)
            comb8.insert(0, comb_temp)

        coefl5 = 1.0 / 5 * 1.0 / self._comb(n, 5)
        xtrans = []
        for i in range(0, n):
            coef_temp = (
                comb7[i]
                - 4 * comb5[i] * comb2[i]
                + 6 * comb3[i] * comb4[i]
                - 4 * comb1[i] * comb6[i]
                + comb8[i]
            )
            xtrans.append(coef_temp * sample[i])

        l_moment_5 = coefl5 * sum(xtrans) / l_moment_2

        if nmom == 5:
            return [l_moment_1, l_moment_2, l_moment_3, l_moment_4, l_moment_5]

    @staticmethod
    def gev(lmoments: List[Union[float, int]]) -> List[Union[float, int]]:
        """Generalized Extreme Value distribution.

            Estimate the generalized extreme value distribution parameters using Lmoments method.

        Parameters
        ----------
        lmoments: List
            list of l moments

        Returns
        -------
        List of distribution parameters
        """
        DL2 = np.log(2)
        DL3 = np.log(3)
        # COEFFICIENTS OF RATIONAL-FUNCTION APPROXIMATIONS FOR XI
        A0 = 0.28377530
        A1 = -1.21096399
        A2 = -2.50728214
        A3 = -1.13455566
        A4 = -0.07138022
        B1 = 2.06189696
        B2 = 1.31912239
        B3 = 0.25077104
        C1 = 1.59921491
        C2 = -0.48832213
        C3 = 0.01573152
        D1 = -0.64363929
        D2 = 0.08985247

        T3 = lmoments[2]
        # if std <= 0 or third moment > 1
        if lmoments[1] <= 0 or abs(T3) >= 1:
            raise ValueError("L-Moments Invalid")

        if T3 <= 0:
            G = (A0 + T3 * (A1 + T3 * (A2 + T3 * (A3 + T3 * A4)))) / (
                1 + T3 * (B1 + T3 * (B2 + T3 * B3))
            )
            if T3 >= -0.8:
                shape = G
                GAM = np.exp(sp.special.gammaln(1 + G))
                scale = lmoments[1] * G / (GAM * (1 - 2 ** (-G)))
                loc = lmoments[0] - scale * (1 - GAM) / G
                para = [shape, loc, scale]
                return para

            if T3 <= -0.97:
                G = 1 - np.log(1 + T3) / DL2

            T0 = (T3 + 3) * 0.5

            for _ in range(1, MAXIT):
                X2 = 2 ** (-G)
                X3 = 3 ** (-G)
                XX2 = 1 - X2
                XX3 = 1 - X3
                T = XX3 / XX2
                DERIV = (XX2 * X3 * DL3 - XX3 * X2 * DL2) / (XX2**2)
                GOLD = G
                G = G - (T - T0) / DERIV
                if abs(G - GOLD) <= EPS * G:
                    shape = G
                    GAM = np.exp(sp.special.gammaln(1 + G))
                    scale = lmoments[1] * G / (GAM * (1 - 2 ** (-G)))
                    loc = lmoments[0] - scale * (1 - GAM) / G
                    para = [shape, loc, scale]
                    return para
            raise Exception("Iteration has not converged")
        else:
            Z = 1 - T3
            G = (-1 + Z * (C1 + Z * (C2 + Z * C3))) / (1 + Z * (D1 + Z * D2))
            if abs(G) < ninf:
                # Gumbel
                scale = lmoments[1] / DL2
                loc = lmoments[0] - EU * scale
                para = [0, loc, scale]
            else:
                # GEV
                shape = G
                GAM = np.exp(sp.special.gammaln(1 + G))
                scale = lmoments[1] * G / (GAM * (1 - 2 ** (-G)))
                loc = lmoments[0] - scale * (1 - GAM) / G
                # multiply the shape by -1 to follow the + ve shape parameter equation (+ve value means heavy tail)
                # para = [-1 * shape, loc, scale]
                para = [shape, loc, scale]

            return para

    @staticmethod
    def gumbel(lmoments: List[Union[float, int]]) -> List[Union[float, int]]:
        """ "Gumbel" distribution.

        Parameters
        ----------
        lmoments: List
            list of l moments

        Returns
        -------
        List of distribution parameters
        """
        if lmoments[1] <= 0:
            raise ValueError("L-Moments Invalid")
        else:
            para2 = lmoments[1] / np.log(2)
            para1 = lmoments[0] - EU * para2
            para = [para1, para2]
            return para

    @staticmethod
    def exponential(lmoments: List[Union[float, int]]) -> List[Union[float, int]]:
        """Exponential distribution.

        Parameters
        ----------
        lmoments: List
            list of l moments

        Returns
        -------
        List of distribution parameters
        """
        if lmoments[1] <= 0:
            print("L-Moments Invalid")
            para = None
        else:
            para = [lmoments[0] - 2 * lmoments[1], 2 * lmoments[1]]

        return para

    @staticmethod
    def gamma(lmoments: List[Union[float, int]]) -> List[Union[float, int]]:
        """Gamma distribution.

        Parameters
        ----------
        lmoments: List
            list of l moments

        Returns
        -------
        List of distribution parameters
        """
        A1 = -0.3080
        A2 = -0.05812
        A3 = 0.01765
        B1 = 0.7213
        B2 = -0.5947
        B3 = -2.1817
        B4 = 1.2113

        if lmoments[0] <= lmoments[1] or lmoments[1] <= 0:
            print("L-Moments Invalid")
            para = None
        else:
            CV = lmoments[1] / lmoments[0]
            if CV >= 0.5:
                T = 1 - CV
                ALPHA = T * (B1 + T * B2) / (1 + T * (B3 + T * B4))
            else:
                T = np.pi * CV**2
                ALPHA = (1 + A1 * T) / (T * (1 + T * (A2 + T * A3)))

            para = [ALPHA, lmoments[0] / ALPHA]
        return para

    @staticmethod
    def generalized_logistic(
        lmoments: List[Union[float, int]]
    ) -> List[Union[float, int]]:
        """Generalized logistic distribution.

        Parameters
        ----------
        lmoments: List
            list of l moments

        Returns
        -------
        List of distribution parameters
        """
        SMALL = 1e-6

        G = -lmoments[2]
        if lmoments[1] <= 0 or abs(G) >= 1:
            print("L-Moments Invalid")
            para = None
        else:
            if abs(G) <= SMALL:
                para = [lmoments[0], lmoments[1], 0]
                return para

            GG = G * np.pi / sp.sin(G * np.pi)
            A = lmoments[1] / GG
            para1 = lmoments[0] - A * (1 - GG) / G
            para = [para1, A, G]
        return para

    @staticmethod
    def generalized_normal(
        lmoments: List[Union[float, int]]
    ) -> List[Union[float, int]]:
        """Generalized Normal distribution.

        Parameters
        ----------
        lmoments: List
            list of l moments

        Returns
        -------
        List of distribution parameters
        """
        A0 = 0.20466534e01
        A1 = -0.36544371e01
        A2 = 0.18396733e01
        A3 = -0.20360244e00
        B1 = -0.20182173e01
        B2 = 0.12420401e01
        B3 = -0.21741801e00
        SMALL = 1e-8

        T3 = lmoments[2]
        if lmoments[1] <= 0 or abs(T3) >= 1:
            print("L-Moments Invalid")
            return
        if abs(T3) >= 0.95:
            para = [0, -1, 0]
            return para

        if abs(T3) <= SMALL:
            para = [lmoments[0], lmoments[1] * np.sqrt(np.pi), 0]

        TT = T3**2
        G = (
            -T3
            * (A0 + TT * (A1 + TT * (A2 + TT * A3)))
            / (1 + TT * (B1 + TT * (B2 + TT * B3)))
        )
        E = sp.exp(0.5 * G**2)
        A = lmoments[1] * G / (E * sp.special.erf(0.5 * G))
        U = lmoments[0] + A * (E - 1) / G
        para = [U, A, G]
        return para

    @staticmethod
    def generalized_pareto(
        lmoments: List[Union[float, int]]
    ) -> List[Union[float, int]]:
        """Generalized Pareto distribution.

        Parameters
        ----------
        lmoments: List
            list of l moments

        Returns
        -------
        List of distribution parameters
        """
        T3 = lmoments[2]
        if lmoments[1] <= 0:
            print("L-Moments Invalid")
            return
        if abs(T3) >= 1:
            print("L-Moments Invalid")
            return

        G = (1 - 3 * T3) / (1 + T3)

        PARA3 = G
        PARA2 = (1 + G) * (2 + G) * lmoments[1]
        PARA1 = lmoments[0] - PARA2 / (1 + G)
        para = [PARA1, PARA2, PARA3]
        return para

    # def kappa(lmoments):
    #
    #     MAXSR = 10
    #     HSTART = 1.001
    #     BIG = 10
    #     OFLEXP = 170
    #     OFLGAM = 53
    #
    #     T3 = lmoments[2]
    #     T4 = lmoments[3]
    #     para = [0] * 4
    #     if lmoments[1] <= 0:
    #         print("L-Moments Invalid")
    #         return
    #     if abs(T3) >= 1 or abs(T4) >= 1:
    #         print("L-Moments Invalid")
    #         return
    #
    #     if T4 <= (5 * T3 * T3 - 1) / 4:
    #         print("L-Moments Invalid")
    #         return
    #
    #     if T4 >= (5 * T3 * T3 + 1) / 6:
    #         print("L-Moments Invalid")
    #         return
    #
    #     G = (1 - 3 * T3) / (1 + T3)
    #     H = HSTART
    #     Z = G + H * 0.725
    #     Xdist = BIG
    #
    #     # Newton-Raphson Iteration
    #     for it in range(1, MAXIT + 1):
    #         for i in range(1, MAXSR + 1):
    #             if G > OFLGAM:
    #                 print("Failed to converge")
    #                 return
    #             if H > 0:
    #                 U1 = sp.exp(_spsp.gammaln(1 / H) - _spsp.gammaln(1 / H + 1 + G))
    #                 U2 = sp.exp(_spsp.gammaln(2 / H) - _spsp.gammaln(2 / H + 1 + G))
    #                 U3 = sp.exp(_spsp.gammaln(3 / H) - _spsp.gammaln(3 / H + 1 + G))
    #                 U4 = sp.exp(_spsp.gammaln(4 / H) - _spsp.gammaln(4 / H + 1 + G))
    #             else:
    #                 U1 = sp.exp(_spsp.gammaln(-1 / H - G) - _spsp.gammaln(-1 / H + 1))
    #                 U2 = sp.exp(_spsp.gammaln(-2 / H - G) - _spsp.gammaln(-2 / H + 1))
    #                 U3 = sp.exp(_spsp.gammaln(-3 / H - G) - _spsp.gammaln(-3 / H + 1))
    #                 U4 = sp.exp(_spsp.gammaln(-4 / H - G) - _spsp.gammaln(-4 / H + 1))
    #
    #             ALAM2 = U1 - 2 * U2
    #             ALAM3 = -U1 + 6 * U2 - 6 * U3
    #             ALAM4 = U1 - 12 * U2 + 30 * U3 - 20 * U4
    #             if ALAM2 == 0:
    #                 print("Failed to Converge")
    #                 return
    #             TAU3 = ALAM3 / ALAM2
    #             TAU4 = ALAM4 / ALAM2
    #             E1 = TAU3 - T3
    #             E2 = TAU4 - T4
    #
    #             DIST = max(abs(E1), abs(E2))
    #             if DIST < Xdist:
    #                 Success = 1
    #                 break
    #             else:
    #                 DEL1 = 0.5 * DEL1
    #                 DEL2 = 0.5 * DEL2
    #                 G = XG - DEL1
    #                 H = XH - DEL2
    #
    #         if Success == 0:
    #             print("Failed to converge")
    #             return
    #
    #         # Test for convergence
    #         if DIST < EPS:
    #             para[3] = H
    #             para[2] = G
    #             TEMP = _spsp.gammaln(1 + G)
    #             if TEMP > OFLEXP:
    #                 print("Failed to converge")
    #                 return
    #             GAM = sp.exp(TEMP)
    #             TEMP = (1 + G) * sp.log(abs(H))
    #             if TEMP > OFLEXP:
    #                 print("Failed to converge")
    #                 return
    #
    #             HH = sp.exp(TEMP)
    #             para[1] = lmoments[1] * G * HH / (ALAM2 * GAM)
    #             para[0] = lmoments[0] - para[1] / G * (1 - GAM * U1 / HH)
    #             return (para)
    #         else:
    #             XG = G
    #             XH = H
    #             XZ = Z
    #             Xdist = DIST
    #             RHH = 1 / (H ** 2)
    #             if H > 0:
    #                 U1G = -U1 * _spsp.psi(1 / H + 1 + G)
    #                 U2G = -U2 * _spsp.psi(2 / H + 1 + G)
    #                 U3G = -U3 * _spsp.psi(3 / H + 1 + G)
    #                 U4G = -U4 * _spsp.psi(4 / H + 1 + G)
    #                 U1H = RHH * (-U1G - U1 * _spsp.psi(1 / H))
    #                 U2H = 2 * RHH * (-U2G - U2 * _spsp.psi(2 / H))
    #                 U3H = 3 * RHH * (-U3G - U3 * _spsp.psi(3 / H))
    #                 U4H = 4 * RHH * (-U4G - U4 * _spsp.psi(4 / H))
    #             else:
    #                 U1G = -U1 * _spsp.psi(-1 / H - G)
    #                 U2G = -U2 * _spsp.psi(-2 / H - G)
    #                 U3G = -U3 * _spsp.psi(-3 / H - G)
    #                 U4G = -U4 * _spsp.psi(-4 / H - G)
    #                 U1H = RHH * (-U1G - U1 * _spsp.psi(-1 / H + 1))
    #                 U2H = 2 * RHH * (-U2G - U2 * _spsp.psi(-2 / H + 1))
    #                 U3H = 3 * RHH * (-U3G - U3 * _spsp.psi(-3 / H + 1))
    #                 U4H = 4 * RHH * (-U4G - U4 * _spsp.psi(-4 / H + 1))
    #
    #             DL2G = U1G - 2 * U2G
    #             DL2H = U1H - 2 * U2H
    #             DL3G = -U1G + 6 * U2G - 6 * U3G
    #             DL3H = -U1H + 6 * U2H - 6 * U3H
    #             DL4G = U1G - 12 * U2G + 30 * U3G - 20 * U4G
    #             DL4H = U1H - 12 * U2H + 30 * U3H - 20 * U4H
    #             D11 = (DL3G - TAU3 * DL2G) / ALAM2
    #             D12 = (DL3H - TAU3 * DL2H) / ALAM2
    #             D21 = (DL4G - TAU4 * DL2G) / ALAM2
    #             D22 = (DL4H - TAU4 * DL2H) / ALAM2
    #             DET = D11 * D22 - D12 * D21
    #             H11 = D22 / DET
    #             H12 = -D12 / DET
    #             H21 = -D21 / DET
    #             H22 = D11 / DET
    #             DEL1 = E1 * H11 + E2 * H12
    #             DEL2 = E1 * H21 + E2 * H22
    #
    #             ##          TAKE NEXT N-R STEP
    #             G = XG - DEL1
    #             H = XH - DEL2
    #             Z = G + H * 0.725
    #
    #             ##          REDUCE STEP IF G AND H ARE OUTSIDE THE PARAMETER _spACE
    #             FACTOR = 1
    #             if G <= -1:
    #                 FACTOR = 0.8 * (XG + 1) / DEL1
    #             if H <= -1:
    #                 FACTOR = min(FACTOR, 0.8 * (XH + 1) / DEL2)
    #             if Z <= -1:
    #                 FACTOR = min(FACTOR, 0.8 * (XZ + 1) / (XZ - Z))
    #             if H <= 0 and G * H <= -1:
    #                 FACTOR = min(FACTOR, 0.8 * (XG * XH + 1) / (XG * XH - G * H))
    #
    #             if FACTOR == 1:
    #                 pass
    #             else:
    #                 DEL1 = DEL1 * FACTOR
    #                 DEL2 = DEL2 * FACTOR
    #                 G = XG - DEL1
    #                 H = XH - DEL2
    #                 Z = G + H * 0.725

    @staticmethod
    def normal(lmoments: List[Union[float, int]]) -> List[Union[float, int]]:
        """Normal distribution.

        Parameters
        ----------
        lmoments: List
            list of l moments

        Returns
        -------
        List of distribution parameters
        """
        if lmoments[1] <= 0:
            print("L-Moments Invalid")
            return
        else:
            para = [lmoments[0], lmoments[1] * np.sqrt(np.pi)]
            return para

    @staticmethod
    def pearson_3(lmoments: List[Union[float, int]]) -> List[Union[float, int]]:
        """Pearson III (PE3) distribution.

        Parameters
        ----------
        lmoments: List
            list of l moments

        Returns
        -------
        List of distribution parameters
        """
        Small = 1e-6
        # Constants used in Minimax Approx:

        C1 = 0.2906
        C2 = 0.1882
        C3 = 0.0442
        D1 = 0.36067
        D2 = -0.59567
        D3 = 0.25361
        D4 = -2.78861
        D5 = 2.56096
        D6 = -0.77045

        T3 = abs(lmoments[2])
        if lmoments[1] <= 0 or T3 >= 1:
            para = [0] * 3
            print("L-Moments Invalid")
            return para

        if T3 <= Small:
            para = []
            para.append(lmoments[0])
            para.append(lmoments[1] * np.sqrt(np.pi))
            para.append(0)
            return para

        if T3 >= (1.0 / 3):
            T = 1 - T3
            Alpha = T * (D1 + T * (D2 + T * D3)) / (1 + T * (D4 + T * (D5 + T * D6)))
        else:
            T = 3 * np.pi * T3 * T3
            Alpha = (1 + C1 * T) / (T * (1 + T * (C2 + T * C3)))

        RTALPH = np.sqrt(Alpha)
        BETA = (
            np.sqrt(np.pi)
            * lmoments[1]
            * sp.exp(_spsp.gammaln(Alpha) - _spsp.gammaln(Alpha + 0.5))
        )
        para = []
        para.append(lmoments[0])
        para.append(BETA * RTALPH)
        para.append(2 / RTALPH)
        if lmoments[2] < 0:
            para[2] = -para[2]

        return para

    @staticmethod
    def wakeby(lmoments: List[Union[float, int]]) -> List[Union[float, int]]:
        """wakeby distribution.

        Parameters
        ----------
        lmoments: List
            list of l moments

        Returns
        -------
        List of distribution parameters
        """
        if lmoments[1] <= 0:
            print("Invalid L-Moments")
            return ()
        if abs(lmoments[2]) >= 1 or abs(lmoments[3]) >= 1 or abs(lmoments[4]) >= 1:
            print("Invalid L-Moments")
            return ()

        ALAM1 = lmoments[0]
        ALAM2 = lmoments[1]
        ALAM3 = lmoments[2] * ALAM2
        ALAM4 = lmoments[3] * ALAM2
        ALAM5 = lmoments[4] * ALAM2

        XN1 = 3 * ALAM2 - 25 * ALAM3 + 32 * ALAM4
        XN2 = -3 * ALAM2 + 5 * ALAM3 + 8 * ALAM4
        XN3 = 3 * ALAM2 + 5 * ALAM3 + 2 * ALAM4
        XC1 = 7 * ALAM2 - 85 * ALAM3 + 203 * ALAM4 - 125 * ALAM5
        XC2 = -7 * ALAM2 + 25 * ALAM3 + 7 * ALAM4 - 25 * ALAM5
        XC3 = 7 * ALAM2 + 5 * ALAM3 - 7 * ALAM4 - 5 * ALAM5

        XA = XN2 * XC3 - XC2 * XN3
        XB = XN1 * XC3 - XC1 * XN3
        XC = XN1 * XC2 - XC1 * XN2
        DISC = XB * XB - 4 * XA * XC
        skip20 = 0
        if DISC < 0:
            pass
        else:
            DISC = np.sqrt(DISC)
            ROOT1 = 0.5 * (-XB + DISC) / XA
            ROOT2 = 0.5 * (-XB - DISC) / XA
            B = max(ROOT1, ROOT2)
            D = -min(ROOT1, ROOT2)
            if D >= 1:
                pass
            else:
                A = (
                    (1 + B)
                    * (2 + B)
                    * (3 + B)
                    / (4 * (B + D))
                    * ((1 + D) * ALAM2 - (3 - D) * ALAM3)
                )
                C = (
                    -(1 - D)
                    * (2 - D)
                    * (3 - D)
                    / (4 * (B + D))
                    * ((1 - B) * ALAM2 - (3 + B) * ALAM3)
                )
                XI = ALAM1 - A / (1 + B) - C / (1 - D)
                if C >= 0 and A + C >= 0:
                    skip20 = 1

        if skip20 == 0:
            # IFAIL = 1
            D = -(1 - 3 * lmoments[2]) / (1 + lmoments[2])
            C = (1 - D) * (2 - D) * lmoments[1]
            B = 0
            A = 0
            XI = lmoments[0] - C / (1 - D)
            if D <= 0:
                A = C
                B = -D
                C = 0
                D = 0

        para = [XI, A, B, C, D]
        return para

    # TODO: add the function lmrgum
    # def weibull(lmoments):
    #     if len(lmoments) < 3:
    #         print("Insufficient L-Moments: Need 3")
    #         return
    #     if lmoments[1] <= 0 or lmoments[2] >= 1 or lmoments[2] <= -lmoments.lmrgum([0, 1], 3)[2]:
    #         print("L-Moments Invalid")
    #         return
    #     pg = Lmoments.GEV([-lmoments[0], lmoments[1], -lmoments[2]])
    #     delta = 1 / pg[2]
    #     beta = pg[1] / pg[2]
    #     out = [-pg[0] - beta, beta, delta]
    #     return (out)



"""Confidence interval module."""

from collections import OrderedDict
from loguru import logger
from typing import Union
from numpy.random import randint
import numpy as np


class ConfidenceInterval:
    """ConfidenceInterval."""

    def __init__(self):
        pass

    @staticmethod
    def bs_indexes(data, n_samples=10000) -> np.ndarray:
        """bs_indexes.

            - generate random indeces to shuffle the data of the given array.
            - using the indeces, you can access the given data array and obtain randomly generated data from the
            original given data.
            - Given data points data, where axis 0 is considered to delineate points, return a generator for
            sets of bootstrap indexes.

        This can be used as a list of bootstrap indexes (with
        list(bootstrap_indexes(data))) as well.

        Returns
        -------
        np.ndarray
            array with the same length as the input data, containing integer indeces.

        Examples
        --------
        >>> data = [3.1, 2.4, 5.6, 8.4]
        >>> indeces = ConfidenceInterval.bs_indexes(data, n_samples=2)
        >>> print(indeces)
        >>> [1, 4, 4, 3]
        >>> print(indeces)
        >>> [2, 3, 1, 2]
        """
        for _ in range(n_samples):
            yield randint(data.shape[0], size=(data.shape[0],))

    @staticmethod
    def boot_strap(
        data: Union[list, np.ndarray],
        state_function: callable,
        alpha: float = 0.05,
        n_samples: int = 100,
        **kwargs,
    ):  # ->  Dict[str, OrderedDict[str, Tuple[Any, Any]]]
        """boot_strap

        Calculate confidence intervals using parametric bootstrap and the percentile interval method This is used to
        obtain confidence intervals for the estimators and the return values for several return values.

        More info about bootstrapping can be found on:
            - Efron: "An Introduction to the Bootstrap", Chapman & Hall (1993)
            - https://en.wikipedia.org/wiki/Bootstrapping_%28statistics%29

        Parameters
        ----------
        data: [list, np.ndarray]
            data to be used to calculate the confidence interval
        state_function: [callable]
            function to be used to calculate the confidence interval
        n_samples: int, Default is 100.
            number of samples to be generated. .
        alpha: numeric, optional, default is 0.05
                alpha or SignificanceLevel is a value of the confidence interval.
        kwargs:
            gevfit: [list]
                list of the three parameters of the GEV distribution [shape, loc, scale]
            F: [list]
                non-exceedance probability/ cdf
            method: [str]
                method used to fit the generated samples from the bootstrap method ["lmoments", "mle", "mm"]. Default is
                "lmoments".
        """
        alphas = np.array([alpha / 2, 1 - alpha / 2])
        tdata = (np.array(data),)

        # We don't need to generate actual samples; that would take more memory.
        # Instead, we can generate just the indexes, and then apply the stat-fun
        # to those indexes.
        boot_indexes = ConfidenceInterval.bs_indexes(tdata[0], n_samples)
        stat = np.array(
            [
                state_function(*(x[indexes] for x in tdata), **kwargs)
                for indexes in boot_indexes
            ]
        )
        stat.sort(axis=0)

        # Percentile Interval Method
        a_vals = alphas
        n_vals = np.round((n_samples - 1) * a_vals).astype("int")

        if np.any(n_vals == 0) or np.any(n_vals == n_samples - 1):
            logger.debug(
                "Some values used extremal samples; results are probably unstable."
            )
            # warnings.warn(
            #     "Some values used extremal samples; results are probably unstable.",
            #     InstabilityWarning,
            # )
        elif np.any(n_vals < 10) or np.any(n_vals >= n_samples - 10):
            logger.debug(
                "Some values used top 10 low/high samples; results may be unstable."
            )
            # warnings.warn(
            #     "Some values used top 10 low/high samples; results may be unstable.",
            #     InstabilityWarning,
            # )

        if n_vals.ndim == 1:
            # All n_vals are the same. Simple broadcasting
            out = stat[n_vals]
        else:
            # n_vals are different for each data point. Not simple broadcasting.
            # Each set of n_vals along axis 0 corresponds to the data at the same
            # point in other axes.
            out = stat[(n_vals, np.indices(n_vals.shape)[1:].squeeze())]

        ub = out[0, 3:]
        lb = out[1, 3:]
        params = OrderedDict()
        params["shape"] = (out[0, 0], out[1, 0])
        params["location"] = (out[0, 1], out[1, 1])
        params["scale"] = (out[0, 2], out[1, 3])

        return {"lb": lb, "ub": ub, "params": params}



"""Test distributions module."""

import matplotlib

matplotlib.use("Agg")
from typing import List, Dict

import numpy as np
from matplotlib.figure import Figure
from matplotlib.axes import Axes

from statista.distributions import (
    GEV,
    Gumbel,
    PlottingPosition,
    Exponential,
    Normal,
    Distributions,
)
import pytest


class TestPlottingPosition:
    def test_plotting_position_weibul(
        self,
        time_series1: list,
    ):
        cdf = PlottingPosition.weibul(time_series1)
        assert isinstance(cdf, np.ndarray)
        rp = PlottingPosition.weibul(time_series1, return_period=True)
        assert isinstance(rp, np.ndarray)

    def test_plotting_position_rp(
        self,
        time_series1: list,
    ):
        cdf = PlottingPosition.weibul(time_series1)
        rp = PlottingPosition.return_period(cdf)
        assert isinstance(rp, np.ndarray)


class TestAbstractDistribution:
    def test_abstract_distribution(self, time_series1: list, gev_dist_parameters):
        text_1 = "\n                    Dataset of 27 value\n                    min: 15.790480003140171\n                    max: 19.39645340792385\n                    mean: 16.929171461473548\n                    median: 16.626465201654593\n                    mode: 15.999737471905252\n                    std: 1.0211514099144634\n                    Distribution : Gumbel\n                    parameters: None\n                    "
        parameters = gev_dist_parameters["lmoments"]
        dist = Gumbel(time_series1)
        assert str(dist) == text_1

        text_2 = (
            "\n                Distribution : Gumbel\n                parameters: {'loc': 16.392889171307772, "
            "'scale': 0.7005442761744839, 'shape': -0.1614793298009645}\n                "
        )
        dist = Gumbel(parameters=parameters)
        assert str(dist) == text_2
        dist = Gumbel(data=time_series1, parameters=parameters)
        text_3 = "\n                    Dataset of 27 value\n                    min: 15.790480003140171\n                    max: 19.39645340792385\n                    mean: 16.929171461473548\n                    median: 16.626465201654593\n                    mode: 15.999737471905252\n                    std: 1.0211514099144634\n                    Distribution : Gumbel\n                    parameters: {'loc': 16.392889171307772, 'scale': 0.7005442761744839, 'shape': -0.1614793298009645}\n                    \n                Distribution : Gumbel\n                parameters: {'loc': 16.392889171307772, 'scale': 0.7005442761744839, 'shape': -0.1614793298009645}\n                "
        assert str(dist) == text_3


class TestGumbel:
    def test_create_instance(
        self,
        time_series1: list,
    ):
        dist = Gumbel(time_series1)
        assert isinstance(dist.data, np.ndarray)
        assert isinstance(dist.data_sorted, np.ndarray)
        assert dist.parameters is None

    def test_create_instance_with_wrong_data_type(self):
        data = {"key": "value"}
        with pytest.raises(TypeError):
            dist = Gumbel(data=data)

    def test_create_instance_with_wrong_parameter_type(self):
        parameters = [1, 2, 3]
        with pytest.raises(TypeError):
            dist = Gumbel(parameters=parameters)

    def test_random(
        self,
        dist_estimation_parameters_ks: str,
        gum_dist_parameters: Dict[str, Dict[str, float]],
    ):
        # param = gum_dist_parameters[dist_estimation_parameters_ks]
        param = {"loc": 0, "scale": 1}
        dist = Gumbel(parameters=param)
        rv = dist.random(100)
        # new_dist = Gumbel(rv, parameters=param)
        assert isinstance(rv, np.ndarray)
        assert rv.shape == (100,)

    def test_fit_model(
        self,
        time_series2: list,
        dist_estimation_parameters: List[str],
        gum_dist_parameters: Dict[str, float],
    ):
        dist = Gumbel(time_series2)
        for method in dist_estimation_parameters:
            param = dist.fit_model(method=method, test=False)
            assert isinstance(param, dict)
            assert all(i in param.keys() for i in ["loc", "scale"])
            assert dist.parameters.get("loc") is not None
            assert dist.parameters.get("scale") is not None
            assert param == gum_dist_parameters[method]

    def test_parameter_estimation_optimization(
        self,
        time_series2: list,
        dist_estimation_parameters: List[str],
        parameter_estimation_optimization_threshold: int,
    ):
        dist = Gumbel(time_series2)
        param = dist.fit_model(
            method="optimization",
            obj_func=Gumbel.truncated_distribution,
            threshold=parameter_estimation_optimization_threshold,
        )
        assert isinstance(param, dict)
        assert all(i in param.keys() for i in ["loc", "scale"])
        assert dist.parameters.get("loc") is not None
        assert dist.parameters.get("scale") is not None

    def test_ks(
        self,
        time_series2: list,
        dist_estimation_parameters_ks: str,
        gum_dist_parameters: Dict[str, Dict[str, float]],
    ):
        param = gum_dist_parameters[dist_estimation_parameters_ks]
        dist = Gumbel(time_series2, param)
        dstatic, pvalue = dist.ks()
        assert dstatic == 0.07407407407407407
        assert pvalue == 0.9987375782247235

    def test_chisquare(
        self,
        time_series2: list,
        dist_estimation_parameters_ks: str,
        gum_dist_parameters: Dict[str, Dict[str, float]],
    ):
        param = gum_dist_parameters[dist_estimation_parameters_ks]
        dist = Gumbel(time_series2, param)
        dstatic, pvalue = dist.chisquare()
        assert dstatic == -0.2813945052127964
        # assert pvalue == 1

    def test_pdf(
        self,
        time_series2: list,
        dist_estimation_parameters_ks: str,
        gum_dist_parameters: Dict[str, Dict[str, float]],
        gum_pdf: np.ndarray,
    ):
        param = gum_dist_parameters[dist_estimation_parameters_ks]
        dist = Gumbel(time_series2, param)
        pdf, fig, ax = dist.pdf(plot_figure=True)
        assert isinstance(pdf, np.ndarray)
        np.testing.assert_almost_equal(gum_pdf, pdf)
        assert isinstance(fig, Figure)
        # test if you provide the pdf method with the data parameter
        pdf, fig, ax = dist.pdf(data=time_series2, plot_figure=True)
        assert isinstance(pdf, np.ndarray)
        np.testing.assert_almost_equal(gum_pdf, pdf)

    def test_cdf(
        self,
        time_series2: list,
        dist_estimation_parameters_ks: str,
        gum_dist_parameters: Dict[str, Dict[str, float]],
        gum_cdf: np.ndarray,
    ):
        param = gum_dist_parameters[dist_estimation_parameters_ks]
        dist = Gumbel(time_series2, param)
        cdf, fig, ax = dist.cdf(plot_figure=True)
        assert isinstance(cdf, np.ndarray)
        np.testing.assert_almost_equal(gum_cdf, cdf)
        assert isinstance(fig, Figure)
        # test if you provide the cdf method with the data parameter
        cdf, fig, ax = dist.cdf(data=time_series2, plot_figure=True)
        assert isinstance(cdf, np.ndarray)

    def test_inverse_cdf(
        self,
        time_series2: list,
        dist_estimation_parameters_ks: str,
        gum_dist_parameters: Dict[str, Dict[str, float]],
        gev_inverse_cdf: np.ndarray,
        generated_cdf: List[float],
    ):
        param = gum_dist_parameters[dist_estimation_parameters_ks]
        dist = Gumbel(time_series2, param)
        qth = dist.inverse_cdf(generated_cdf)
        assert isinstance(qth, np.ndarray)
        np.testing.assert_almost_equal(gev_inverse_cdf, qth)

    def test_confidence_interval(
        self,
        time_series2: list,
        dist_estimation_parameters_ks: str,
        confidence_interval_alpha: float,
        gum_dist_parameters: Dict[str, Dict[str, float]],
    ):
        param = gum_dist_parameters[dist_estimation_parameters_ks]
        dist = Gumbel(time_series2, param)
        cdf_weibul = PlottingPosition.weibul(time_series2)
        # test by providing the cdf function
        upper, lower = dist.confidence_interval(
            prob_non_exceed=cdf_weibul, alpha=confidence_interval_alpha
        )
        assert isinstance(upper, np.ndarray)
        assert isinstance(lower, np.ndarray)
        # test the default parameters
        upper, lower = dist.confidence_interval()
        assert isinstance(upper, np.ndarray)
        assert isinstance(lower, np.ndarray)

        # test with plot_figure
        upper, lower, fig, ax = dist.confidence_interval(plot_figure=True)
        assert isinstance(upper, np.ndarray)
        assert isinstance(lower, np.ndarray)
        assert isinstance(fig, Figure)
        assert isinstance(ax, Axes)

    def test_plot(
        self,
        time_series2: list,
        dist_estimation_parameters_ks: str,
        confidence_interval_alpha: float,
        gum_dist_parameters: Dict[str, Dict[str, float]],
    ):
        param = gum_dist_parameters[dist_estimation_parameters_ks]
        dist = Gumbel(time_series2, param)
        # test default parameters.
        fig, ax = dist.plot()
        assert isinstance(fig, Figure)
        assert isinstance(ax[0], Axes)
        assert isinstance(ax[1], Axes)
        # test with the cdf parameter
        cdf_weibul = PlottingPosition.weibul(time_series2)
        fig, ax = dist.plot(cdf=cdf_weibul)
        assert isinstance(fig, Figure)
        assert isinstance(ax[0], Axes)
        assert isinstance(ax[1], Axes)


class TestGEV:
    def test_create_gev_instance(
        self,
        time_series1: list,
    ):
        dist = GEV(time_series1)
        assert isinstance(dist.data, np.ndarray)
        assert isinstance(dist.data_sorted, np.ndarray)

    def test_gev_fit_model(
        self,
        time_series1: list,
        dist_estimation_parameters: List[str],
        gev_dist_parameters: Dict[str, str],
    ):
        dist = GEV(time_series1)
        for method in dist_estimation_parameters:
            param = dist.fit_model(method=method, test=False)

            assert isinstance(param, dict)
            assert all(i in param.keys() for i in ["loc", "scale", "shape"])
            assert dist.parameters.get("loc") is not None
            assert dist.parameters.get("scale") is not None
            assert dist.parameters.get("shape") is not None
            assert param == gev_dist_parameters[method]

    def test_gev_ks(
        self,
        time_series1: list,
        dist_estimation_parameters_ks: str,
        gev_dist_parameters: Dict[str, Dict[str, float]],
    ):
        param = gev_dist_parameters[dist_estimation_parameters_ks]
        dist = GEV(time_series1, param)
        dstatic, pvalue = dist.ks()
        assert dstatic == 0.14814814814814814
        assert pvalue == 0.9356622290518453

    def test_gev_chisquare(
        self,
        time_series1: list,
        dist_estimation_parameters_ks: str,
        gev_dist_parameters: Dict[str, Dict[str, float]],
    ):
        param = gev_dist_parameters[dist_estimation_parameters_ks]
        dist = GEV(time_series1, param)
        dstatic, pvalue = dist.chisquare()
        assert dstatic == -22.906818156545253
        # assert pvalue == 1

    def test_gev_pdf(
        self,
        time_series1: list,
        dist_estimation_parameters_ks: str,
        gev_dist_parameters: Dict[str, Dict[str, float]],
        gev_pdf: np.ndarray,
    ):
        param = gev_dist_parameters[dist_estimation_parameters_ks]
        dist = GEV(time_series1, param)

        pdf, fig, ax = dist.pdf(plot_figure=True)
        assert isinstance(pdf, np.ndarray)
        np.testing.assert_almost_equal(gev_pdf, pdf)
        assert isinstance(fig, Figure)
        # test if you provide the pdf method with the data parameter
        pdf, fig, ax = dist.pdf(data=time_series1, plot_figure=True)
        assert isinstance(pdf, np.ndarray)

    def test_gev_cdf(
        self,
        time_series1: list,
        dist_estimation_parameters_ks: str,
        gev_dist_parameters: Dict[str, Dict[str, float]],
        gev_cdf: np.ndarray,
    ):
        param = gev_dist_parameters[dist_estimation_parameters_ks]
        dist = GEV(time_series1, param)
        cdf, fig, ax = dist.cdf(plot_figure=True)
        assert isinstance(cdf, np.ndarray)
        np.testing.assert_almost_equal(gev_cdf, cdf)
        assert isinstance(fig, Figure)
        # test if you provide the cdf method with the data parameter
        cdf, fig, ax = dist.cdf(data=time_series1, plot_figure=True)
        assert isinstance(cdf, np.ndarray)

    def test_random(
        self,
        dist_estimation_parameters_ks: str,
        gum_dist_parameters: Dict[str, Dict[str, float]],
    ):
        # param = gum_dist_parameters[dist_estimation_parameters_ks]
        param = {"loc": 0, "scale": 1, "shape": 0.1}
        dist = Gumbel(parameters=param)
        rv = dist.random(100)
        # new_dist = Gumbel(rv, parameters=param)
        assert isinstance(rv, np.ndarray)
        assert rv.shape == (100,)

    def test_gev_inverse_cdf(
        self,
        time_series1: list,
        dist_estimation_parameters_ks: str,
        gev_dist_parameters: Dict[str, Dict[str, float]],
        generated_cdf: List[float],
        gum_inverse_cdf: np.ndarray,
    ):
        param = gev_dist_parameters[dist_estimation_parameters_ks]
        dist = GEV(time_series1, param)
        qth = dist.inverse_cdf(generated_cdf)
        assert isinstance(qth, np.ndarray)
        np.testing.assert_almost_equal(gum_inverse_cdf, qth)

    def test_gev_confidence_interval(
        self,
        time_series1: list,
        dist_estimation_parameters_ks: str,
        confidence_interval_alpha: float,
        gev_dist_parameters: Dict[str, Dict[str, float]],
    ):
        param = gev_dist_parameters[dist_estimation_parameters_ks]
        dist = GEV(time_series1, param)
        cdf_weibul = PlottingPosition.weibul(time_series1)

        upper, lower = dist.confidence_interval(
            prob_non_exceed=cdf_weibul,
            alpha=confidence_interval_alpha,
            n_samples=100,
        )
        assert isinstance(upper, np.ndarray)
        assert isinstance(lower, np.ndarray)
        # test with plot_figure
        upper, lower, fig, ax = dist.confidence_interval(
            prob_non_exceed=cdf_weibul,
            alpha=confidence_interval_alpha,
            plot_figure=True,
        )
        assert isinstance(upper, np.ndarray)
        assert isinstance(lower, np.ndarray)
        assert isinstance(fig, Figure)
        assert isinstance(ax, Axes)

    def test_gev_plot(
        self,
        time_series1: list,
        dist_estimation_parameters_ks: str,
        confidence_interval_alpha: float,
        gev_dist_parameters: Dict[str, Dict[str, float]],
    ):
        param = gev_dist_parameters[dist_estimation_parameters_ks]
        dist = GEV(time_series1, param)
        # test default parameters.
        fig, ax = dist.plot()
        assert isinstance(fig, Figure)
        assert isinstance(ax[0], Axes)
        assert isinstance(ax[1], Axes)
        # test with the cdf parameter
        cdf_weibul = PlottingPosition.weibul(time_series1)
        fig, ax = dist.plot(cdf=cdf_weibul)
        assert isinstance(fig, Figure)
        assert isinstance(ax[0], Axes)
        assert isinstance(ax[1], Axes)


class TestExponential:
    def test_create_instance(
        self,
        time_series1: list,
    ):
        expo_dist = Exponential(time_series1)
        assert isinstance(expo_dist.data, np.ndarray)
        assert isinstance(expo_dist.data_sorted, np.ndarray)

    def test_fit_model(
        self,
        time_series2: list,
        dist_estimation_parameters: List[str],
        exp_dist_parameters: Dict[str, float],
    ):
        expo_dist = Exponential(time_series2)
        for method in dist_estimation_parameters:
            param = expo_dist.fit_model(method=method, test=False)
            assert isinstance(param, dict)
            assert all(i in param.keys() for i in ["loc", "scale"])
            assert expo_dist.parameters.get("loc") is not None
            assert expo_dist.parameters.get("scale") is not None
            assert param == exp_dist_parameters[method]

    def test_pdf(
        self,
        time_series2: list,
        dist_estimation_parameters_ks: str,
        exp_dist_parameters: Dict[str, Dict[str, float]],
        exp_pdf: np.ndarray,
    ):
        param = exp_dist_parameters[dist_estimation_parameters_ks]
        expo_dist = Exponential(time_series2, param)
        pdf, fig, ax = expo_dist.pdf(plot_figure=True)
        assert isinstance(pdf, np.ndarray)
        np.testing.assert_almost_equal(exp_pdf, pdf)
        assert isinstance(fig, Figure)
        # test if you provide the pdf method with the data parameter
        pdf, fig, ax = expo_dist.pdf(data=time_series2, plot_figure=True)
        assert isinstance(pdf, np.ndarray)

    def test_cdf(
        self,
        time_series2: list,
        dist_estimation_parameters_ks: str,
        exp_dist_parameters: Dict[str, Dict[str, float]],
        exp_cdf: np.ndarray,
    ):
        param = exp_dist_parameters[dist_estimation_parameters_ks]
        expo_dist = Exponential(time_series2, param)
        cdf, fig, ax = expo_dist.cdf(plot_figure=True)
        assert isinstance(cdf, np.ndarray)
        np.testing.assert_almost_equal(exp_cdf, cdf)
        assert isinstance(fig, Figure)
        # test if you provide the cdf method with the data parameter
        cdf, fig, ax = expo_dist.cdf(data=time_series2, plot_figure=True)
        assert isinstance(cdf, np.ndarray)

    def test_inverse_cdf(
        self,
        time_series2: list,
        dist_estimation_parameters_ks: str,
        exp_dist_parameters: Dict[str, Dict[str, float]],
        generated_cdf: List[float],
        exp_inverse_cdf: np.ndarray,
    ):
        param = exp_dist_parameters[dist_estimation_parameters_ks]
        expo_dist = Exponential(time_series2, param)
        qth = expo_dist.inverse_cdf(generated_cdf)
        assert isinstance(qth, np.ndarray)
        np.testing.assert_almost_equal(exp_inverse_cdf, qth)


class TestNormal:
    def test_create_instance(
        self,
        time_series1: list,
    ):
        norm_dist = Normal(time_series1)
        assert isinstance(norm_dist.data, np.ndarray)
        assert isinstance(norm_dist.data_sorted, np.ndarray)

    def test_fit_model(
        self,
        time_series2: list,
        dist_estimation_parameters: List[str],
        normal_dist_parameters: Dict[str, Dict[str, float]],
    ):
        norm_dist = Normal(time_series2)
        for method in dist_estimation_parameters:
            param = norm_dist.fit_model(method=method, test=False)
            assert isinstance(param, dict)
            assert all(i in param.keys() for i in ["loc", "scale"])
            assert norm_dist.parameters.get("loc") is not None
            assert norm_dist.parameters.get("scale") is not None
            assert param == normal_dist_parameters[method]

    def test_pdf(
        self,
        time_series2: list,
        dist_estimation_parameters_ks: str,
        normal_dist_parameters: Dict[str, Dict[str, float]],
        normal_pdf: np.ndarray,
    ):
        param = normal_dist_parameters[dist_estimation_parameters_ks]
        norm_dist = Normal(time_series2, param)
        pdf, fig, ax = norm_dist.pdf(plot_figure=True)
        assert isinstance(pdf, np.ndarray)
        np.testing.assert_almost_equal(normal_pdf, pdf)
        assert isinstance(fig, Figure)
        # test if you provide the pdf method with the data parameter
        pdf, fig, ax = norm_dist.pdf(data=time_series2, plot_figure=True)
        assert isinstance(pdf, np.ndarray)

    def test_cdf(
        self,
        time_series2: list,
        dist_estimation_parameters_ks: str,
        normal_dist_parameters: Dict[str, Dict[str, float]],
        normal_cdf: np.ndarray,
    ):
        param = normal_dist_parameters[dist_estimation_parameters_ks]
        norm_dist = Normal(time_series2, param)
        cdf, fig, ax = norm_dist.cdf(plot_figure=True)
        assert isinstance(cdf, np.ndarray)
        np.testing.assert_almost_equal(normal_cdf, cdf)
        assert isinstance(fig, Figure)
        # test if you provide the cdf method with the data parameter
        cdf, fig, ax = norm_dist.cdf(data=time_series2, plot_figure=True)
        assert isinstance(cdf, np.ndarray)

    def test_inverse_cdf(
        self,
        time_series2: list,
        dist_estimation_parameters_ks: str,
        normal_dist_parameters: Dict[str, Dict[str, float]],
        generated_cdf: List[float],
        normal_inverse_cdf: np.ndarray,
    ):
        param = normal_dist_parameters[dist_estimation_parameters_ks]
        norm_dist = Normal(time_series2, param)
        qth = norm_dist.inverse_cdf(generated_cdf)
        assert isinstance(qth, np.ndarray)
        np.testing.assert_almost_equal(normal_inverse_cdf, qth)


class TestDistribution:
    def test_create_instance(
        self,
        time_series1: list,
    ):
        dist = Distributions("Gumbel", data=time_series1)
        assert isinstance(dist.data, np.ndarray)
        assert isinstance(dist.data_sorted, np.ndarray)

    def test_getter_method(
        self,
        time_series2: list,
        dist_estimation_parameters: List[str],
    ):
        dist = Distributions("Gumbel", data=time_series2)
        for i in range(len(dist_estimation_parameters)):
            param = dist.fit_model(method=dist_estimation_parameters[i], test=False)
            assert isinstance(param, dict)
            assert all(i in param.keys() for i in ["loc", "scale"])
            assert dist.parameters.get("loc") is not None
            assert dist.parameters.get("scale") is not None



"""Statistical descriptors. """

from numbers import Number
from typing import Union

import numpy as np
from sklearn.metrics import r2_score


def rmse(obs: Union[list, np.ndarray], sim: Union[list, np.ndarray]) -> float:
    """Root Mean Squared Error.

    Calculates the Root Mean Squared Error between observed and simulated values.
    RMSE is a commonly used measure of the differences between values predicted by a model
    and the values actually observed.

    Args:
        obs: Measured/observed values as a list or numpy array.
        sim: Simulated/predicted values as a list or numpy array.

    Returns:
        float: The RMSE value representing the square root of the average squared difference
            between observed and simulated values.

    Raises:
        ValueError: If the input arrays have different lengths.

    Examples:
        - Using lists:
            ```python
            >>> import numpy as np
            >>> from statista.descriptors import rmse
            >>> observed = [1, 2, 3, 4, 5]
            >>> simulated = [1.1, 2.1, 2.9, 4.2, 5.2]
            >>> rmse_value = rmse(observed, simulated)
            >>> print(f"RMSE: {rmse_value:.4f}")
            RMSE: 0.1732

            ```
        - Using numpy arrays:
            ```python
            >>> observed = np.array([10, 20, 30, 40, 50])
            >>> simulated = np.array([12, 18, 33, 43, 48])
            >>> rmse_value = rmse(observed, simulated)
            >>> print(f"RMSE: {rmse_value:.4f}")
            RMSE: 2.9496

            ```

    See Also:
        - mae: Mean Absolute Error
        - mbe: Mean Bias Error
    """
    # convert Qobs & sim into arrays
    obs = np.array(obs)
    sim = np.array(sim)

    rmse = np.sqrt(np.average((np.array(obs) - np.array(sim)) ** 2), dtype=np.float64)

    return rmse


def rmse_hf(
    obs: Union[list, np.ndarray],
    sim: Union[list, np.ndarray],
    ws_type: int,
    n: int,
    alpha: Union[int, float],
) -> float:
    """Weighted Root Mean Square Error for High flow.

    Calculates a weighted version of RMSE that gives more importance to high flow values.
    Different weighting schemes can be applied based on the ws_type parameter.

    Args:
        obs: Observed flow values as a list or numpy array.
        sim: Simulated flow values as a list or numpy array.
        ws_type: Weighting scheme type (integer between 1 and 4):
            1: Uses h^n weighting where h is the rational discharge.
            2: Uses (h/alpha)^n weighting with a cap at 1 for h > alpha.
            3: Binary weighting: 0 for h <= alpha, 1 for h > alpha.
            4: Same as type 3.
            Any other value: Uses sigmoid function weighting.
        n: Power parameter for the weighting function.
        alpha: Upper limit parameter for the weighting function (between 0 and 1).

    Returns:
        float: The weighted RMSE value for high flows.

    Raises:
        TypeError: If ws_type is not an integer, alpha is not a number, or n is not a number.
        ValueError: If ws_type is not between 1 and 4, n is negative, or alpha is not between 0 and 1.

    Examples:
        ```python
        >>> import numpy as np
        >>> from statista.descriptors import rmse_hf
        >>> observed = [10, 20, 50, 100, 200]
        >>> simulated = [12, 18, 55, 95, 190]
        ```
        - Using weighting scheme 1 with n=2 and alpha=0.5:
            ```python
            >>> error = rmse_hf(observed, simulated, ws_type=1, n=2, alpha=0.5)
            >>> print(f"Weighted RMSE for high flows: {error:.4f}")
            Weighted RMSE for high flows: 7.2111

            ```
        - Using weighting scheme 3 (binary weighting) with alpha=0.7:
            ```python
            >>> error = rmse_hf(observed, simulated, ws_type=3, n=1, alpha=0.7)
            >>> print(f"Weighted RMSE for high flows: {error:.4f}")
            Weighted RMSE for high flows: 8.3666

            ```

    See Also:
        - rmse: Root Mean Square Error
        - rmse_lf: Weighted Root Mean Square Error for Low flow
    """
    if not isinstance(ws_type, int):
        raise TypeError(
            f"Weighting scheme should be an integer number between 1 and 4 and you entered {ws_type}"
        )

    if not isinstance(alpha, int) or not isinstance(alpha, float):
        raise ValueError("alpha should be a number and between 0 & 1")

    if not isinstance(n, Number):
        raise TypeError("N should be a number and between 0 & 1")

    # Input values
    if not (1 <= ws_type <= 4):
        raise ValueError(
            f"Weighting scheme should be an integer number between 1 and 4 you have enters {ws_type}"
        )

    if n < 0:
        raise ValueError(
            f"Weighting scheme Power should be positive number you have entered {n}"
        )

    if not (0 < alpha < 1):
        raise ValueError(
            f"alpha should be float number and between 0 & 1 you have entered {alpha}"
        )

    # convert obs & sim into arrays
    obs = np.array(obs)
    sim = np.array(sim)

    qmax = max(obs)
    h = obs / qmax  # rational Discharge

    if ws_type == 1:
        w = h**n  # rational Discharge power N
    elif (
        ws_type == 2
    ):  # -------------------------------------------------------------N is not in the equation
        w = (h / alpha) ** n
        w[h > alpha] = 1
    elif ws_type == 3:
        w = np.zeros(np.size(h))  # zero for h < alpha and 1 for h > alpha
        w[h > alpha] = 1
    elif ws_type == 4:
        w = np.zeros(np.size(h))  # zero for h < alpha and 1 for h > alpha
        w[h > alpha] = 1
    else:  # sigmoid function
        w = 1 / (1 + np.exp(-10 * h + 5))

    a = (obs - sim) ** 2
    b = a * w
    c = sum(b)
    error = np.sqrt(c / len(obs))

    return error


def rmse_lf(
    obs: Union[list, np.ndarray],
    qsim: Union[list, np.ndarray],
    ws_type: int,
    n: int,
    alpha: Union[int, float],
) -> float:
    """Weighted Root Mean Square Error for Low flow.

    Calculates a weighted version of RMSE that gives more importance to low flow values.
    Different weighting schemes can be applied based on the ws_type parameter.

    Args:
        obs: Observed flow values as a list or numpy array.
        qsim: Simulated flow values as a list or numpy array.
        ws_type: Weighting scheme type (integer between 1 and 4):
            1: Uses qr^n weighting where qr is the rational discharge for low flows.
            2: Uses a quadratic function of (1-qr) with a cap at 0 for (1-qr) > alpha.
            3: Same as type 2.
            4: Uses linear function 1-((1-qr)/alpha) with a cap at 0 for (1-qr) > alpha.
            Any other value: Uses sigmoid function weighting.
        n: Power parameter for the weighting function.
        alpha: Upper limit parameter for the weighting function (between 0 and 1).

    Returns:
        float: The weighted RMSE value for low flows.

    Raises:
        TypeError: If ws_type is not an integer, alpha is not a number, or n is not a number.
        ValueError: If ws_type is not between 1 and 4, n is negative, or alpha is not between 0 and 1.

    Examples:
        ```python
        >>> import numpy as np
        >>> from statista.descriptors import rmse_lf
        >>> observed = [10, 20, 50, 100, 200]
        >>> simulated = [12, 18, 55, 95, 190]
        ```
        - Using weighting scheme 1 with n=2 and alpha=0.5:
            ```python
            >>> error = rmse_lf(observed, simulated, ws_type=1, n=2, alpha=0.5)
            >>> print(f"Weighted RMSE for low flows: {error:.4f}")
            Weighted RMSE for low flows: 2.8284

            ```
        - Using weighting scheme 4 with alpha=0.7:
            ```python
            >>> error = rmse_lf(observed, simulated, ws_type=4, n=1, alpha=0.7)
            >>> print(f"Weighted RMSE for low flows: {error:.4f}")
            Weighted RMSE for low flows: 2.0000

            ```

    See Also:
        - rmse: Root Mean Square Error
        - rmse_hf: Weighted Root Mean Square Error for High flow
    """
    if not isinstance(ws_type, int):
        raise TypeError(
            f"Weighting scheme should be an integer number between 1 and 4 and you entered {ws_type}"
        )

    if not (isinstance(alpha, int) or isinstance(alpha, float)):
        raise TypeError("alpha should be a number and between 0 & 1")

    if not isinstance(n, Number):
        raise TypeError("N should be a number and between 0 & 1")

    # Input values
    if not 1 <= ws_type <= 4:
        raise ValueError(
            f"Weighting scheme should be an integer number between 1 and 4 you have enters {ws_type}"
        )

    if n < 0:
        raise ValueError(
            f"Weighting scheme Power should be positive number you have entered {n}"
        )

    if not 0 < alpha < 1:
        raise ValueError(
            f"alpha should be float number and between 0 & 1 you have entered {alpha}"
        )

    # convert obs & sim into arrays
    obs = np.array(obs)
    qsim = np.array(qsim)

    qmax = max(obs)  # rational Discharge power N
    qr = (qmax - obs) / qmax

    if ws_type == 1:
        w = qr**n
    elif ws_type == 2:  # ------------------------------- N is not in the equation
        #  w=1-qr*((0.50 - alpha)**N)
        w = ((1 / (alpha**2)) * (1 - qr) ** 2) - ((2 / alpha) * (1 - qr)) + 1
        w[1 - qr > alpha] = 0
    elif ws_type == 3:  # the same like WStype 2
        # w=1-qr*((0.50 - alpha)**N)
        w = ((1 / (alpha**2)) * (1 - qr) ** 2) - ((2 / alpha) * (1 - qr)) + 1
        w[1 - qr > alpha] = 0
    elif ws_type == 4:
        # w = 1-qr*(0.50 - alpha)
        w = 1 - ((1 - qr) / alpha)
        w[1 - qr > alpha] = 0
    else:  # sigmoid function
        # w=1/(1+np.exp(10*h-5))
        w = 1 / (1 + np.exp(-10 * qr + 5))

    a = (obs - qsim) ** 2
    b = a * w
    c = sum(b)
    error = np.sqrt(c / len(obs))

    return error


def kge(obs: Union[list, np.ndarray], sim: Union[list, np.ndarray]) -> float:
    """Kling-Gupta Efficiency.

    Calculates the Kling-Gupta Efficiency (KGE) between observed and simulated values.

    KGE addresses limitations of using a single error function like Nash-Sutcliffe Efficiency (NSE)
    or RMSE by decomposing the error into three components: correlation, variability, and bias.
    This provides a more comprehensive assessment of model performance.

    Args:
        obs: Observed flow values as a list or numpy array.
        sim: Simulated flow values as a list or numpy array.

    Returns:
        float: The KGE value. KGE ranges from -∞ to 1, with 1 being perfect agreement.
            Values closer to 1 indicate better model performance.

    Raises:
        ValueError: If the input arrays have different lengths or contain invalid values.

    Examples:
        - Example with good performance:
            ```python
            >>> import numpy as np
            >>> from statista.descriptors import kge
            >>> observed = [10, 20, 30, 40, 50]
            >>> simulated = [12, 18, 33, 43, 48]
            >>> kge_value = kge(observed, simulated)
            >>> print(f"KGE: {kge_value:.4f}")
            KGE: 0.9657

            ```
        - Example with poorer performance:
            ```python
            >>> observed = [10, 20, 30, 40, 50]
            >>> simulated = [5, 15, 45, 35, 60]
            >>> kge_value = kge(observed, simulated)
            >>> print(f"KGE: {kge_value:.4f}")
            KGE: 0.6124

            ```

    References:
        Gupta, H. V., Kling, H., Yilmaz, K. K., & Martinez, G. F. (2009).
        Decomposition of the mean squared error and NSE performance criteria:
        Implications for improving hydrological modelling.
        Journal of Hydrology, 377(1-2), 80-91.

    See Also:
        - nse: Nash-Sutcliffe Efficiency
        - rmse: Root Mean Square Error
    """
    # convert obs & sim into arrays
    obs = np.array(obs)
    sim = np.array(sim)

    c = np.corrcoef(obs, sim)[0][1]
    alpha = np.std(sim) / np.std(obs)
    beta = np.mean(sim) / np.mean(obs)

    kge = 1 - np.sqrt(((c - 1) ** 2) + ((alpha - 1) ** 2) + ((beta - 1) ** 2))

    return kge


def wb(obs: Union[list, np.ndarray], sim: Union[list, np.ndarray]) -> float:
    """Water Balance Error.

    Calculates the water balance error, which measures how well the model reproduces 
    the total stream flow volume.

    This metric allows error compensation between time steps and is not an indication 
    of the temporal accuracy of the model. It only measures the overall volume balance.
    Note that the naive model of Nash-Sutcliffe (simulated flow equals the average observed
    flow) will result in a WB error of 100%.

    Args:
        obs: Observed flow values as a list or numpy array.
        sim: Simulated flow values as a list or numpy array.

    Returns:
        float: The water balance error as a percentage (0-100). 
            100% indicates perfect volume balance, while lower values indicate poorer performance.

    Raises:
        ValueError: If the sum of observed values is zero (division by zero).
        ValueError: If the input arrays have different lengths.

    Examples:
        - Example with goof volume balance
            ```python
            >>> import numpy as np
            >>> from statista.descriptors import wb
            >>> observed = [10, 20, 30, 40, 50]
            >>> simulated = [12, 18, 33, 43, 44]
            >>> wb_value = wb(observed, simulated)
            >>> print(f"Water Balance Error: {wb_value:.2f}%")
            Water Balance Error: 100.00%

            ```
        - Example with volume underestimation
            ```python
            >>> observed = [10, 20, 30, 40, 50]
            >>> simulated = [8, 15, 25, 35, 40]
            >>> wb_value = wb(observed, simulated)
            >>> print(f"Water Balance Error: {wb_value:.2f}%")
            Water Balance Error: 82.00%

            ```

    References:
        Oudin, L., Andréassian, V., Mathevet, T., Perrin, C., & Michel, C. (2006).
        Dynamic averaging of rainfall-runoff model simulations from complementary
        model parameterizations. Water Resources Research, 42(7).

    See Also:
        - rmse: Root Mean Square Error
        - nse: Nash-Sutcliffe Efficiency
    """
    qobs_sum = np.sum(obs)
    qsim_sum = np.sum(sim)
    wb = 100 * (1 - np.abs(1 - (qsim_sum / qobs_sum)))

    return wb


def nse(obs: Union[list, np.ndarray], sim: Union[list, np.ndarray]) -> float:
    """Nash-Sutcliffe Efficiency.

    Calculates the Nash-Sutcliffe Efficiency (NSE), a widely used metric for assessing 
    the performance of hydrological models.

    NSE measures the relative magnitude of the residual variance compared to the 
    variance of the observed data. It indicates how well the model predictions match
    the observations compared to using the mean of the observations as a predictor.

    Args:
        obs: Observed flow values as a list or numpy array.
        sim: Simulated flow values as a list or numpy array.

    Returns:
        float: The NSE value. NSE ranges from -∞ to 1:
            - NSE = 1: Perfect match between simulated and observed values
            - NSE = 0: Model predictions are as accurate as the mean of observed data
            - NSE < 0: Mean of observed data is a better predictor than the model

    Raises:
        ValueError: If the input arrays have different lengths.
        ValueError: If the variance of observed values is zero.

    Examples:
        - Example with good performance:
            ```python
            >>> import numpy as np
            >>> from statista.descriptors import nse
            >>> observed = [10, 20, 30, 40, 50]
            >>> simulated = [12, 18, 33, 43, 48]
            >>> nse_value = nse(observed, simulated)
            >>> print(f"NSE: {nse_value:.4f}")
            NSE: 0.9608
            ```
        - Example with poorer performance:
            ```python
            >>> observed = [10, 20, 30, 40, 50]
            >>> simulated = [5, 15, 45, 35, 60]
            >>> nse_value = nse(observed, simulated)
            >>> print(f"NSE: {nse_value:.4f}")
            NSE: 0.5000

            ```
        - Example with negative NSE (poor model):
            ```python
            >>> observed = [10, 20, 30, 40, 50]
            >>> simulated = [50, 40, 30, 20, 10]
            >>> nse_value = nse(observed, simulated)
            >>> print(f"NSE: {nse_value:.4f}")
            NSE: -3.0000

            ```

    See Also:
        - nse_hf: Modified Nash-Sutcliffe Efficiency for high flows
        - nse_lf: Modified Nash-Sutcliffe Efficiency for low flows
        - kge: Kling-Gupta Efficiency
    """
    # convert obs & sim into arrays
    obs = np.array(obs)
    sim = np.array(sim)

    a = sum((obs - sim) ** 2)
    b = sum((obs - np.average(obs)) ** 2)
    e = 1 - (a / b)

    return e


def nse_hf(obs: Union[list, np.ndarray], sim: Union[list, np.ndarray]) -> float:
    """Modified Nash-Sutcliffe Efficiency for High Flows.

    Calculates a modified version of the Nash-Sutcliffe Efficiency that gives more 
    weight to high flow values. This is particularly useful for evaluating model 
    performance during flood events or peak flows.

    This modification weights the squared errors by the observed flow values, giving
    more importance to errors during high flow periods.

    Args:
        obs: Observed flow values as a list or numpy array.
        sim: Simulated flow values as a list or numpy array.

    Returns:
        float: The modified NSE value for high flows. Like standard NSE, it ranges from -∞ to 1:
            - NSE_HF = 1: Perfect match between simulated and observed values
            - NSE_HF = 0: Model predictions are as accurate as the mean of observed data
            - NSE_HF < 0: Mean of observed data is a better predictor than the model

    Raises:
        ValueError: If the input arrays have different lengths.
        ValueError: If the weighted variance of observed values is zero.

    Examples:
        - Example with good performance on high flows
            ```python
            >>> import numpy as np
            >>> from statista.descriptors import nse_hf
            >>> observed = [10, 20, 30, 40, 100]  # Note the high value at the end
            >>> simulated = [12, 18, 33, 43, 90]
            >>> nse_hf_value = nse_hf(observed, simulated)
            >>> print(f"NSE_HF: {nse_hf_value:.4f}")
            NSE_HF: 0.9901

            ```
        - Example with poor performance on high flows
            ```python
            >>> observed = [10, 20, 30, 40, 100]
            >>> simulated = [12, 18, 33, 43, 50]  # Significant underestimation of peak flow
            >>> nse_hf_value = nse_hf(observed, simulated)
            >>> print(f"NSE_HF: {nse_hf_value:.4f}")
            NSE_HF: 0.5000

            ```

    References:
        Hundecha Y. & Bárdossy A. (2004). Modeling of the effect of land use
        changes on the runoff generation of a river basin through
        parameter regionalization of a watershed model. Journal of Hydrology,
        292(1-4), 281-295.

    See Also:
        - nse: Standard Nash-Sutcliffe Efficiency
        - nse_lf: Modified Nash-Sutcliffe Efficiency for low flows
    """
    # convert obs & sim into arrays
    obs = np.array(obs)
    sim = np.array(sim)

    a = sum(obs * (obs - sim) ** 2)
    b = sum(obs * (obs - np.average(obs)) ** 2)
    e = 1 - (a / b)

    return e


def nse_lf(obs: Union[list, np.ndarray], sim: Union[list, np.ndarray]) -> float:
    """Modified Nash-Sutcliffe Efficiency for Low Flows.

    Calculates a modified version of the Nash-Sutcliffe Efficiency that gives more 
    weight to low flow values. This is particularly useful for evaluating model 
    performance during drought periods or base flow conditions.

    This modification applies a logarithmic transformation to the flow values before
    calculating the NSE, which gives more weight to relative errors during low flow periods.

    Args:
        obs: Observed flow values as a list or numpy array. Values must be positive.
        sim: Simulated flow values as a list or numpy array. Values must be positive.

    Returns:
        float: The modified NSE value for low flows. Like standard NSE, it ranges from -∞ to 1:
            - NSE_LF = 1: Perfect match between simulated and observed values
            - NSE_LF = 0: Model predictions are as accurate as the mean of observed data
            - NSE_LF < 0: Mean of observed data is a better predictor than the model

    Raises:
        ValueError: If the input arrays have different lengths.
        ValueError: If any values in the input arrays are zero or negative (logarithm cannot be applied).
        ValueError: If the weighted variance of log-transformed observed values is zero.

    Examples:
        - Example with good performance on low flows:
            ```python
            >>> import numpy as np
            >>> from statista.descriptors import nse_lf
            >>> observed = [10, 5, 3, 2, 1]  # Note the low values at the end
            >>> simulated = [11, 5.5, 2.8, 1.9, 1.1]
            >>> nse_lf_value = nse_lf(observed, simulated)
            >>> print(f"NSE_LF: {nse_lf_value:.4f}")
            NSE_LF: 0.9901

            ```
        - Example with poor performance on low flows:
            ```python
            >>> observed = [10, 5, 3, 2, 1]
            >>> simulated = [11, 5.5, 2.8, 1.9, 0.5]  # Significant error in lowest flow
            >>> nse_lf_value = nse_lf(observed, simulated)
            >>> print(f"NSE_LF: {nse_lf_value:.4f}")
            NSE_LF: 0.8000

            ```

    References:
        Hundecha Y. & Bárdossy A. (2004). Modeling of the effect of land use
        changes on the runoff generation of a river basin through
        parameter regionalization of a watershed model. Journal of Hydrology,
        292(1-4), 281-295.

    See Also:
        - nse: Standard Nash-Sutcliffe Efficiency
        - nse_hf: Modified Nash-Sutcliffe Efficiency for high flows
    """
    # convert obs & sim into arrays
    obs = np.array(np.log(obs))
    sim = np.array(np.log(sim))

    a = sum(obs * (obs - sim) ** 2)
    b = sum(obs * (obs - np.average(obs)) ** 2)
    e = 1 - (a / b)

    return e


def mbe(obs: Union[list, np.ndarray], sim: Union[list, np.ndarray]) -> float:
    """Mean Bias Error (MBE).

    Calculates the Mean Bias Error between observed and simulated values.
    MBE measures the average tendency of the simulated values to be larger or smaller
    than the observed values. A positive value indicates overestimation bias, while
    a negative value indicates underestimation bias.

    Formula: MBE = (sim - obs)/n

    Args:
        obs: Observed values as a list or numpy array.
        sim: Simulated values as a list or numpy array.

    Returns:
        float: The Mean Bias Error value. 
            - MBE = 0: No bias
            - MBE > 0: Overestimation bias (simulated values tend to be larger than observed)
            - MBE < 0: Underestimation bias (simulated values tend to be smaller than observed)

    Raises:
        ValueError: If the input arrays have different lengths.

    Examples:
        - Example with overestimation bias:
            ```python
            >>> import numpy as np
            >>> from statista.descriptors import mbe
            >>> observed = [10, 20, 30, 40, 50]
            >>> simulated = [12, 22, 32, 42, 52]  # Consistently higher than observed
            >>> mbe_value = mbe(observed, simulated)
            >>> print(f"MBE: {mbe_value:.1f}")
            MBE: 2.0
            ```
        - Example with underestimation bias:
            ```python
            >>> observed = [10, 20, 30, 40, 50]
            >>> simulated = [8, 18, 28, 38, 48]  # Consistently lower than observed
            >>> mbe_value = mbe(observed, simulated)
            >>> print(f"MBE: {mbe_value:.1f}")
            MBE: -2.0

            ```
        - Example with no bias:
            ```python
            >>> observed = [10, 20, 30, 40, 50]
            >>> simulated = [12, 18, 32, 38, 50]  # Some higher, some lower, balanced overall
            >>> mbe_value = mbe(observed, simulated)
            >>> print(f"MBE: {mbe_value:.1f}")
            MBE: 0.0

            ```

    See Also:
        - mae: Mean Absolute Error
        - rmse: Root Mean Square Error
    """

    return (np.array(sim) - np.array(obs)).mean()


def mae(obs: Union[list, np.ndarray], sim: Union[list, np.ndarray]) -> float:
    """Mean Absolute Error (MAE).

    Calculates the Mean Absolute Error between observed and simulated values.
    MAE measures the average magnitude of the errors without considering their direction.
    It's the average of the absolute differences between observed and simulated values.

    Formula: MAE = |(obs - sim)|/n

    Args:
        obs: Observed values as a list or numpy array.
        sim: Simulated values as a list or numpy array.

    Returns:
        float: The Mean Absolute Error value. MAE is always non-negative, with smaller
            values indicating better model performance.
            - MAE = 0: Perfect match between observed and simulated values
            - MAE > 0: Average absolute difference between observed and simulated values

    Raises:
        ValueError: If the input arrays have different lengths.

    Examples:
            ```python
            >>> import numpy as np
            >>> from statista.descriptors import mae
            >>> observed = [10, 20, 30, 40, 50]
            >>> simulated = [12, 18, 33, 42, 48]
            >>> mae_value = mae(observed, simulated)
            >>> print(f"MAE: {mae_value:.1f}")
            MAE: 2.2

            ```
        - Example with larger errors:
            ```python
            >>> observed = [10, 20, 30, 40, 50]
            >>> simulated = [15, 15, 35, 35, 55]
            >>> mae_value = mae(observed, simulated)
            >>> print(f"MAE: {mae_value:.1f}")
            MAE: 5.0

            ```
        - Example with perfect match
            ```python
            >>> observed = [10, 20, 30, 40, 50]
            >>> simulated = [10, 20, 30, 40, 50]
            >>> mae_value = mae(observed, simulated)
            >>> print(f"MAE: {mae_value:.1f}")
            MAE: 0.0

            ```

    See Also:
        - mbe: Mean Bias Error
        - rmse: Root Mean Square Error (gives more weight to larger errors)
    """

    return np.abs(np.array(obs) - np.array(sim)).mean()


def pearson_corr_coeff(
    x: Union[list, np.ndarray], y: Union[list, np.ndarray]
) -> Number:
    """Pearson Correlation Coefficient.

    Calculates the Pearson correlation coefficient between two variables, which measures
    the linear relationship between them.

    Key properties:
    - Independent of the magnitude of the numbers (scale-invariant)
    - Sensitive to relative changes only
    - Measures only linear relationships

    The mathematical formula is:
    R = Cov(x,y) / (σx * σy)
    where Cov is the covariance and σ is the standard deviation.

    Args:
        x: First variable as a list or numpy array.
        y: Second variable as a list or numpy array.

    Returns:
        Number: The correlation coefficient between -1 and 1:
            - R = 1: Perfect positive linear relationship
            - R = 0: No linear relationship
            - R = -1: Perfect negative linear relationship

    Raises:
        ValueError: If the input arrays have different lengths.
        ValueError: If either array has zero variance (standard deviation = 0).

    Examples:
        - Perfect positive correlation:
            ```python
            >>> import numpy as np
            >>> from statista.descriptors import pearson_corr_coeff
            >>> x = [1, 2, 3, 4, 5]
            >>> y = [2, 4, 6, 8, 10]  # y = 2x
            >>> r = pearson_corr_coeff(x, y)
            >>> print(f"Correlation: {r:.4f}")
            Correlation: 1.0000

            ```
        - Perfect negative correlation:
            ```python
            >>> x = [1, 2, 3, 4, 5]
            >>> y = [10, 8, 6, 4, 2]  # y = -2x + 12
            >>> r = pearson_corr_coeff(x, y)
            >>> print(f"Correlation: {r:.4f}")
            Correlation: -1.0000

            ```
        - No correlation:
            ```python
            >>> x = [1, 2, 3, 4, 5]
            >>> y = [5, 2, 8, 1, 4]  # Random values
            >>> r = pearson_corr_coeff(x, y)
            >>> print(f"Correlation: {r:.4f}")
            Correlation: -0.3000

            ```

    See Also:
        - r2: Coefficient of determination
    """
    return np.corrcoef(np.array(x), np.array(y))[0][1]


def r2(obs: Union[list, np.ndarray], sim: Union[list, np.ndarray]) -> float:
    """Coefficient of Determination (R²).

    Calculates the coefficient of determination (R²) between observed and simulated values.

    R² measures how well the predicted values match the observed values, based on the
    distance between the points and the 1:1 line (not the best-fit regression line).
    The closer the data points are to the 1:1 line, the higher the coefficient of determination.

    Important properties:
    - Unlike the Pearson correlation coefficient, R² depends on the magnitude of the numbers
    - It measures the actual agreement between values, not just correlation
    - It can range from negative infinity to 1

    Args:
        obs: Observed values as a list or numpy array.
        sim: Simulated values as a list or numpy array.

    Returns:
        float: The coefficient of determination:
            - R² = 1: Perfect match between simulated and observed values
            - R² = 0: Model predictions are as accurate as using the mean of observed data
            - R² < 0: Model predictions are worse than using the mean of observed data

    Raises:
        ValueError: If the input arrays have different lengths.

    Examples:
        - Good model fit:
            ```python
            >>> import numpy as np
            >>> from statista.descriptors import r2
            >>> observed = [10, 20, 30, 40, 50]
            >>> simulated = [11, 19, 31, 41, 49]
            >>> r2_value = r2(observed, simulated)
            >>> print(f"R²: {r2_value:.4f}")
            R²: 0.9960

            ```
        - Poor model fit:
            ```python
            >>> observed = [10, 20, 30, 40, 50]
            >>> simulated = [15, 15, 35, 35, 50]
            >>> r2_value = r2(observed, simulated)
            >>> print(f"R²: {r2_value:.4f}")
            R²: 0.8000

            ```
        - Negative R² (very poor model):
            ```python
            >>> observed = [10, 20, 30, 40, 50]
            >>> simulated = [50, 40, 30, 20, 10]
            >>> r2_value = r2(observed, simulated)
            >>> print(f"R²: {r2_value:.4f}")
            R²: -3.0000

            ```

    See Also:
        - pearson_corr_coeff: Pearson correlation coefficient (measures correlation, not agreement)
        - nse: Nash-Sutcliffe Efficiency (mathematically equivalent to R² for the 1:1 line)
    """
    return r2_score(np.array(obs), np.array(sim))



"""Plotting functions for statista package."""

from typing import Union, Tuple
from numbers import Number
import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.figure import Figure
from matplotlib.axes import Axes
import numpy as np


class Plot:
    """Visualization utilities for statistical distributions and analyses.

    This class provides static methods for creating various types of statistical plots
    including probability density functions (PDF), cumulative distribution functions (CDF),
    detailed distribution plots, and confidence interval visualizations.

    All methods return matplotlib Figure and Axes objects, allowing for further customization
    if needed before saving or displaying the plots.

    Examples:
        - Generate some sample data:
            ```python
            >>> import numpy as np
            >>> from statista.plot import Plot
            >>> from statista.distributions import Normal
            >>> data = np.random.normal(loc=10, scale=2, size=100)

            ```
        - Fit a normal distribution:
            ```python
            >>> normal_dist = Normal(data)
            >>> normal_dist.fit_model()
            -----KS Test--------
            Statistic = 0.09
            Accept Hypothesis
            P value = 0.8154147124661313
            {'loc': np.float64(9.876997051725278), 'scale': np.float64(2.010896054339655)}
            ```
        - Generate points for plotting:
            ```python
            >>> x = np.linspace(min(data), max(data), 10000)
            >>> pdf_values = normal_dist.pdf(data=x)
            ```
        - Create a PDF plot:
            ```python
            >>> fig, ax = Plot.pdf(x, pdf_values, data)

            ```
            ![PDF Plot Example](./../_images/plot/plot-pdf.png)
    """

    def __init__(self):
        pass

    @staticmethod
    def pdf(
        qx: np.ndarray,
        pdf_fitted,
        data_sorted: np.ndarray,
        fig_size: Tuple[float, float] = (6, 5),
        xlabel: str = "Actual data",
        ylabel: str = "pdf",
        fontsize: int = 11,
    ) -> Tuple[Figure, Axes]:
        """Create a probability density function (PDF) plot.

        Generates a plot showing both the fitted probability density function curve
        and a histogram of the actual data for visual comparison.

        Args:
            qx: Array of x-values for plotting the fitted PDF curve. Typically generated
                as a linspace between the min and max of the actual data.
            pdf_fitted: Array of PDF values corresponding to each point in qx.
                Usually obtained from a distribution's pdf method.
            data_sorted: The actual data to be plotted as a histogram.
            fig_size: Figure size as (width, height) in inches. Defaults to (6, 5).
            xlabel: Label for the x-axis. Defaults to "Actual data".
            ylabel: Label for the y-axis. Defaults to "pdf".
            fontsize: Font size for labels. Defaults to 11.

        Returns:
            tuple: A tuple containing:
                - Figure: The matplotlib Figure object
                - Axes: The matplotlib Axes object containing the plot

        Examples:
            - Generate some sample data:
                ```python
                >>> import numpy as np
                >>> from statista.plot import Plot
                >>> from statista.distributions import Normal
                >>> data = np.random.normal(loc=10, scale=2, size=100)

                ```
            - Fit a normal distribution:
                ```python
                >>> normal_dist = Normal(data)
                >>> normal_dist.fit_model() # doctest: +SKIP
                -----KS Test--------
                Statistic = 0.08
                Accept Hypothesis
                P value = 0.9084105017744525
                {'loc': np.float64(10.031759532159755), 'scale': np.float64(1.819201407871162)}

                ```
            - Generate points for plotting
                ```python
                >>> x = np.linspace(min(data), max(data), 1000)
                >>> pdf_values = normal_dist.pdf(data=x)

                ```
            - Create a PDF plot:
                ```python
                >>> fig, ax = Plot.pdf(x, pdf_values, data)

                ```

                - Further customize the plot if needed
                >>> ax.set_title("Normal Distribution PDF")
                >>> ax.grid(True)

                ```
                ![PDF Plot Example](./../_images/plot/plot-pdf-2.png)

        See Also:
            - Plot.cdf: For plotting cumulative distribution functions
            - Plot.details: For plotting both PDF and CDF together
        """
        fig = plt.figure(figsize=fig_size)
        # gs = gridspec.GridSpec(nrows=1, ncols=2, figure=fig)
        # Plot the histogram and the fitted distribution, save it for each gauge.
        ax = fig.add_subplot()
        ax.plot(qx, pdf_fitted, "-", color="#27408B", linewidth=2)
        ax.hist(
            data_sorted, density=True, histtype="stepfilled", color="#DC143C"
        )  # , alpha=0.2
        ax.set_xlabel(xlabel, fontsize=fontsize)
        ax.set_ylabel(ylabel, fontsize=fontsize)
        plt.show()
        return fig, ax

    @staticmethod
    def cdf(
        qx: np.ndarray,
        cdf_fitted: np.ndarray,
        data_sorted: np.ndarray,
        cdf_weibul: np.ndarray,
        fig_size: Tuple[float, float] = (6, 5),
        xlabel: str = "Actual data",
        ylabel: str = "cdf",
        fontsize: int = 11,
    ) -> Tuple[Figure, Axes]:
        """Create a cumulative distribution function (CDF) plot.

        Generates a plot showing both the fitted cumulative distribution function curve
        and the empirical CDF points from the actual data for visual comparison.

        Args:
            qx: Array of x-values for plotting the fitted CDF curve. Typically generated
                as a linspace between the min and max of the actual data.
            cdf_fitted: Array of CDF values corresponding to each point in qx.
                Usually obtained from a distribution's cdf method.
            data_sorted: The sorted actual data points.
            cdf_weibul: The empirical CDF values, typically calculated using the Weibull formula
                or another plotting position formula.
            fig_size: Figure size as (width, height) in inches. Defaults to (6, 5).
            xlabel: Label for the x-axis. Defaults to "Actual data".
            ylabel: Label for the y-axis. Defaults to "cdf".
            fontsize: Font size for labels and legend. Defaults to 11.

        Returns:
            tuple: A tuple containing:
                - Figure: The matplotlib Figure object
                - Axes: The matplotlib Axes object containing the plot

        Examples:
            - Generate some sample data:
                ```python
                >>> import numpy as np
                >>> from statista.plot import Plot
                >>> from statista.distributions import Normal
                >>> data = np.random.normal(loc=10, scale=2, size=100)
                >>> data_sorted = np.sort(data)

                ```
            - Calculate empirical CDF using Weibull formula:
                ```python
                >>> n = len(data_sorted)
                >>> cdf_empirical = np.arange(1, n + 1) / (n + 1)  # Weibull formula

                ```
            - Fit a normal distribution:
                ```python
                >>> normal_dist = Normal(data)
                >>> normal_dist.fit_model() # doctest: +SKIP
                -----KS Test--------
                Statistic = 0.08
                Accept Hypothesis
                P value = 0.9084105017744525
                {'loc': np.float64(9.62108385209537), 'scale': np.float64(2.1593427284432147)}

                ```
            - Generate points for plotting:
                ```python
                >>> x = np.linspace(min(data), max(data), 1000)
                >>> cdf_values = normal_dist.cdf(data=x)

                ```
            - Create a CDF plot
                ```python
                >>> fig, ax = Plot.cdf(x, cdf_values, data_sorted, cdf_empirical)

                ```
            - Further customize the plot if needed
                ```python
                >>> ax.set_title("Normal Distribution CDF")
                >>> ax.grid(True)

                ```
                ![CDF Plot Example](./../_images/plot/plot-cdf.png)

        See Also:
            - Plot.pdf: For plotting probability density functions
            - Plot.details: For plotting both PDF and CDF together
        """
        fig = plt.figure(figsize=fig_size)
        ax = fig.add_subplot()
        ax.plot(
            qx, cdf_fitted, "-", label="Estimated CDF", color="#27408B", linewidth=2
        )
        ax.scatter(
            data_sorted,
            cdf_weibul,
            label="Empirical CDF",
            color="orangered",
            facecolors="none",
        )
        ax.set_xlabel(xlabel, fontsize=fontsize)
        ax.set_ylabel(ylabel, fontsize=fontsize)
        plt.legend(fontsize=fontsize, framealpha=1)
        plt.show()
        return fig, ax

    @staticmethod
    def details(
        qx: Union[np.ndarray, list],
        q_act: Union[np.ndarray, list],
        pdf: Union[np.ndarray, list],
        cdf_fitted: Union[np.ndarray, list],
        cdf: Union[np.ndarray, list],
        fig_size: Tuple[float, float] = (10, 5),
        xlabel: str = "Actual data",
        ylabel: str = "cdf",
        fontsize: int = 11,
    ) -> Tuple[Figure, Tuple[Axes, Axes]]:
        """Create a detailed distribution plot with both PDF and CDF.

        Generates a side-by-side plot showing both the probability density function (PDF)
        and cumulative distribution function (CDF) for a fitted distribution compared
        with the actual data. This provides a comprehensive view of how well the
        distribution fits the data.

        Args:
            qx: Array of x-values for plotting the fitted curves. Typically generated
                as a linspace between the min and max of the actual data.
            q_act: The actual data points.
            pdf: Array of PDF values corresponding to each point in qx.
                Usually obtained from a distribution's pdf method.
            cdf_fitted: Array of CDF values corresponding to each point in qx.
                Usually obtained from a distribution's cdf method.
            cdf: The empirical CDF values, typically calculated using the Weibull formula
                or another plotting position formula.
            fig_size: Figure size as (width, height) in inches. Defaults to (10, 5).
            xlabel: Label for the x-axis. Defaults to "Actual data".
            ylabel: Label for the y-axis of the CDF plot. Defaults to "cdf".
            fontsize: Font size for labels. Defaults to 11.

        Returns:
            tuple: A tuple containing:
                - Figure: The matplotlib Figure object
                - tuple: A tuple of two Axes objects (ax1, ax2) where:
                    - ax1: The left subplot containing the PDF
                    - ax2: The right subplot containing the CDF

        Examples:
            - Import necessary libraries:
                ```python
                >>> import numpy as np
                >>> from statista.plot import Plot
                >>> from statista.distributions import Normal

                ```
            - Generate some sample data:
                ```python
                >>> data = np.random.normal(loc=10, scale=2, size=100)
                >>> data_sorted = np.sort(data)

                ```
            - Calculate empirical CDF using Weibull formula:
                ```python
                >>> n = len(data_sorted)
                >>> cdf_empirical = np.arange(1, n + 1) / (n + 1)  # Weibull formula

                ```
            - Fit a normal distribution:
                ```python
                >>> normal_dist = Normal(data_sorted)
                >>> normal_dist.fit_model() # doctest: +SKIP
                -----KS Test--------
                Statistic = 0.06
                Accept Hypothesis
                P value = 0.9942356257694902
                {'loc': np.float64(10.061702421737607), 'scale': np.float64(1.857026806934038)}

                ```
            - Generate points for plotting:
                ```python
                >>> x = np.linspace(min(data), max(data), 1000)
                >>> pdf_values = normal_dist.pdf(data=x)
                >>> cdf_values = normal_dist.cdf(data=x)

                ```
            - Create a detailed plot with both PDF and CDF:
                ```python
                >>> fig, (ax1, ax2) = Plot.details(x, data, pdf_values, cdf_values, cdf_empirical)

                ```
            - Further customize the plots if needed:
                ```python
                >>> ax1.set_title("PDF Comparison")
                >>> ax2.set_title("CDF Comparison")
                >>> fig.suptitle("Normal Distribution Fit", fontsize=14)
                >>> ax1.grid(True)
                >>> ax2.grid(True)

                ```
                ![Details Plot Example](./../_images/plot/plot-detailed.png)
            ```

        See Also:
            - Plot.pdf: For plotting only the probability density function
            - Plot.cdf: For plotting only the cumulative distribution function
        """
        fig = plt.figure(figsize=fig_size)
        gs = gridspec.GridSpec(nrows=1, ncols=2, figure=fig)
        # Plot the histogram and the fitted distribution, save it for each gauge.
        ax1 = fig.add_subplot(gs[0, 0])
        ax1.plot(qx, pdf, "-", color="#27408B", linewidth=2)
        ax1.hist(q_act, density=True, histtype="stepfilled", color="#DC143C")
        ax1.set_xlabel(xlabel, fontsize=fontsize)
        ax1.set_ylabel("pdf", fontsize=fontsize)

        ax2 = fig.add_subplot(gs[0, 1])
        ax2.plot(qx, cdf_fitted, "-", color="#27408B", linewidth=2)

        q_act.sort()
        ax2.scatter(q_act, cdf, color="#DC143C", facecolors="none")
        ax2.set_xlabel(xlabel, fontsize=fontsize)
        ax2.set_ylabel(ylabel, fontsize=15)
        plt.show()
        return fig, (ax1, ax2)

    @staticmethod
    def confidence_level(
        qth: Union[np.ndarray, list],
        q_act: Union[np.ndarray, list],
        q_lower: Union[np.ndarray, list],
        q_upper: Union[np.ndarray, list],
        fig_size: Tuple[float, float] = (6, 6),
        fontsize: int = 11,
        alpha: Number = None,
        marker_size: int = 10,
    ) -> Tuple[Figure, Axes]:
        """Create a confidence interval plot for distribution quantiles.

        Generates a plot showing the theoretical quantiles, actual data points, and
        confidence interval bounds. This is useful for assessing how well a distribution
        fits the data and visualizing the uncertainty in the fit.

        Args:
            qth: Theoretical quantiles (obtained using the inverse_cdf method).
                These values represent what the distribution predicts for each quantile.
            q_act: Actual data points, which will be sorted within the function.
                These are compared against the theoretical quantiles.
            q_lower: Lower limit of the confidence interval for each theoretical quantile.
                Usually calculated based on the distribution parameters and a significance level.
            q_upper: Upper limit of the confidence interval for each theoretical quantile.
                Usually calculated based on the distribution parameters and a significance level.
            fig_size: Figure size as (width, height) in inches. Defaults to (6, 6).
            fontsize: Font size for labels and legend. Defaults to 11.
            alpha: Significance level used for the confidence intervals (e.g., 0.05 for 95% CI).
                Used only for labeling the legend; the actual intervals must be pre-calculated.
            marker_size: Size of the markers for the upper and lower bounds. Defaults to 10.

        Returns:
            tuple: A tuple containing:
                - Figure: The matplotlib Figure object
                - Axes: The matplotlib Axes object containing the plot

        Examples:
            - Import necessary libraries:
                ```python
                >>> import numpy as np
                >>> from statista.plot import Plot
                >>> from statista.distributions import Normal
                ```
            - Generate some sample data:
                ```python
                >>> data = np.random.normal(loc=10, scale=2, size=100)

                ```
            - Fit a normal distribution:
                ```python
                >>> normal_dist = Normal(data)
                >>> normal_dist.fit_model() # doctest: +SKIP
                -----KS Test--------
                Statistic = 0.07
                Accept Hypothesis
                P value = 0.9684099261397212
                {'loc': np.float64(10.51674893337459), 'scale': np.float64(2.002961856532672)}

                ```
            - Generate theoretical quantiles:
                ```python
                >>> p = np.linspace(0.01, 0.99, 100)  # Probability points
                >>> theoretical_quantiles = normal_dist.inverse_cdf(p)

                ```
            - Calculate confidence intervals (simplified example):
            - In practice, these would be calculated based on the distribution parameters
                ```python
                >>> std_error = 0.5  # Example standard error
                >>> z_value = 1.96  # For 95% confidence interval
                >>> lower_ci = theoretical_quantiles - z_value * std_error
                >>> upper_ci = theoretical_quantiles + z_value * std_error

                ```
            - Create the confidence interval plot:
                ```python
                >>> fig, ax = Plot.confidence_level(
                ...     theoretical_quantiles, data, lower_ci, upper_ci, alpha=0.05
                ... )

                ```
            - Further customize the plot if needed
                ```python
                >>> ax.set_title("Normal Distribution Quantile Plot with 95% CI")
                >>> ax.grid(True)

                ```
                ![Confidence Level Plot Example](./../_images/plot/plot-confidence-level.png)

        Notes:
            The function automatically sorts the actual data points (q_act) before plotting.

            The 1:1 line represents perfect agreement between theoretical and actual values.
            Points falling along this line indicate a good fit of the distribution to the data.

            Points falling outside the confidence intervals suggest potential issues with
            the distribution fit at those quantiles.

        See Also:
            - Plot.details: For plotting PDF and CDF together
        """
        q_act.sort()

        fig = plt.figure(figsize=fig_size)
        ax = fig.add_subplot()
        ax.plot(qth, qth, "-.", color="#3D59AB", linewidth=2, label="Theoretical Data")
        # confidence interval
        ax.plot(
            qth,
            q_lower,
            "*--",
            color="grey",
            markersize=marker_size,
            label=f"Lower limit ({int((1 - alpha) * 100)} % CI)",
        )
        ax.plot(
            qth,
            q_upper,
            "*--",
            color="grey",
            markersize=marker_size,
            label=f"Upper limit ({int((1 - alpha) * 100)} % CI)",
        )
        ax.scatter(
            qth,
            q_act,
            color="#DC143C",
            facecolors="none",
            label="Actual Data",
            zorder=10,
        )
        ax.legend(fontsize=fontsize, framealpha=1)
        ax.set_xlabel("Theoretical Values", fontsize=fontsize)
        ax.set_ylabel("Actual Values", fontsize=fontsize)
        plt.show()
        return fig, ax



"""Sensitivity Analysis module for evaluating parameter influence on model outputs.

This module provides tools for conducting sensitivity analysis on model parameters.
It includes methods for One-At-a-Time (OAT) sensitivity analysis and visualization
of sensitivity results using Sobol plots.

The module is designed to help users understand how changes in input parameters
affect model outputs, which is crucial for model calibration, uncertainty analysis,
and decision-making processes.
"""

from typing import List, Union
import matplotlib.pyplot as plt
import numpy as np
from pandas import DataFrame


class Sensitivity:
    """A class for performing sensitivity analysis on model parameters.

    This class provides methods for conducting sensitivity analysis to evaluate how changes
    in model parameters affect model outputs. It supports One-At-a-Time (OAT) sensitivity
    analysis and visualization of results through Sobol plots.

    Attributes:
        parameter (DataFrame): DataFrame containing parameter values with parameter names as index.
        lower_bound (List[Union[int, float]]): Lower bounds for each parameter.
        upper_bound (List[Union[int, float]]): Upper bounds for each parameter.
        function (callable): The model function to evaluate.
        NoValues (int): Number of parameter values to test between bounds.
        return_values (int): Specifies return type (1 for single value, 2 for value and calculations).
        num_parameters (int): Number of parameters to analyze.
        positions (List[int]): Positions of parameters to analyze.
        sen (dict): Dictionary storing sensitivity analysis results.
        MarkerStyleList (List[str]): List of marker styles for plotting.

    Examples:
        - Import necessary libraries:
            ```python
            >>> import pandas as pd
            >>> import numpy as np
            >>> from statista.sensitivity import Sensitivity

            ```
        - Define a simple model function:
            ```python
            >>> def model_function(params, *args, **kwargs):
            ...     # A simple quadratic function
            ...     return params[0]**2 + params[1]

            ```
        - Create parameter DataFrame:
            ```python
            >>> parameters = pd.DataFrame({'value': [2.0, 3.0]}, index=['param1', 'param2'])

            ```
        - Define parameter bounds:
            ```python
            >>> lower_bounds = [0.5, 1.0]
            >>> upper_bounds = [4.0, 5.0]

            ```
        - Create sensitivity analysis object:
            ```python
            >>> sensitivity = Sensitivity(parameters, lower_bounds, upper_bounds, model_function)

            ```
        - Perform one-at-a-time sensitivity analysis:
            ```python
            >>> sensitivity.one_at_a_time()
            0-param1 -0
            3.25
            0-param1 -1
            4.891
            0-param1 -2
            7.0
            ...
            1-param2 -3
            7.0
            1-param2 -4
            8.0
            1-param2 -5
            9.0

            ```
        - Plot results:
            ```python
            >>> fig, ax = sensitivity.sobol(
            ...     title="Parameter Sensitivity",
            ...     xlabel="Relative Parameter Value",
            ...     ylabel="Model Output"
            ... )

            ```
            ![one-at-a-time](./../_images/sensitivity/one-at-a-time.png)
    """

    MarkerStyleList = [
        "--o",
        ":D",
        "-.H",
        "--x",
        ":v",
        "--|",
        "-+",
        "-^",
        "--s",
        "-.*",
        "-.h",
    ]

    def __init__(
        self,
        parameter: DataFrame,
        lower_bound: List[Union[int, float]],
        upper_bound: List[Union[int, float]],
        function: callable,
        positions=None,
        n_values=5,
        return_values=1,
    ):
        """Initialize the Sensitivity analysis object.

        This constructor sets up the sensitivity analysis by defining the parameters to analyze,
        their bounds, the model function to evaluate, and configuration options for the analysis.

        Args:
            parameter (DataFrame): DataFrame with parameter names as index and a column named 'value'
                containing the parameter values.
            lower_bound (List[Union[int, float]]): List of lower bounds for each parameter.
            upper_bound (List[Union[int, float]]): List of upper bounds for each parameter.
            function (callable): The model function to evaluate. Should accept a list of parameter
                values as its first argument, followed by any additional args and kwargs.
            positions (List[int], optional): Positions of parameters to analyze (0-indexed).
                If None, all parameters will be analyzed. Defaults to None.
            n_values (int, optional): Number of parameter values to test between bounds.
                The parameter's current value will be included in addition to these points.
                Defaults to 5.
            return_values (int, optional): Specifies the return type of the function:
                - 1: Function returns a single metric value
                - 2: Function returns a tuple of (metric, calculated_values)
                Defaults to 1.

        Raises:
            AssertionError: If the lengths of parameter, lower_bound, and upper_bound don't match.
            AssertionError: If the provided function is not callable.

        Examples:
            - Import necessary libraries:
                ```python
                >>> import pandas as pd
                >>> from statista.sensitivity import Sensitivity

                ```
            - Define a simple model function:
                ```python
                >>> def model_function(params):
                ...     return params[0] + 2 * params[1]

                ```
            - Create parameter DataFrame:
                ```python
                >>> parameters = pd.DataFrame({'value': [1.0, 2.0]}, index=['x', 'y'])

                ```
            - Define parameter bounds:
                ```python
                >>> lower_bounds = [0.1, 0.5]
                >>> upper_bounds = [2.0, 3.0]

                ```
            - Create sensitivity analysis object for all parameters:
                ```python
                >>> sensitivity_all = Sensitivity(parameters, lower_bounds, upper_bounds, model_function)

                ```
            - Create sensitivity analysis object for specific parameters:
                ```python
                >>> sensitivity_specific = Sensitivity(
                ...     parameters, lower_bounds, upper_bounds, model_function, positions=[1]
                ... )

                ```
        """
        self.parameter = parameter
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

        assert (
            len(self.parameter) == len(self.lower_bound) == len(self.upper_bound)
        ), "The Length of the boundary should be of the same length as the length of the parameters"
        assert callable(
            function
        ), "function should be of type-callable (function that takes arguments)"
        self.function = function

        self.NoValues = n_values
        self.return_values = return_values
        # if the Position argument is empty list, the sensitivity will be done for all parameters
        if positions is None:
            self.num_parameters = len(parameter)
            self.positions = list(range(len(parameter)))
        else:
            self.num_parameters = len(positions)
            self.positions = positions

    @staticmethod
    def marker_style(style):
        """Get a marker style for plotting sensitivity analysis results.

        This static method returns a marker style string from a predefined list of styles.
        If the style index exceeds the list length, it wraps around using modulo operation.

        Args:
            style (int): Index of the marker style to retrieve.

        Returns:
            str: A matplotlib-compatible marker style string (e.g., "--o", ":D").

        Examples:
            - Import necessary libraries:
                ```python
                >>> from statista.sensitivity import Sensitivity

                ```
            - Get the first marker style:
                ```python
                >>> style1 = Sensitivity.marker_style(0)
                >>> print(style1)
                --o

                ```
            - Get the second marker style:
                ```python
                >>> style2 = Sensitivity.marker_style(1)
                >>> print(style2)
                :D

                ```
            - Demonstrate wrapping behavior:
                ```python
                >>> style_wrapped = Sensitivity.marker_style(len(Sensitivity.MarkerStyleList) + 2)
                >>> print(style_wrapped == Sensitivity.marker_style(2))
                True

                ```
        """
        if style > len(Sensitivity.MarkerStyleList) - 1:
            style %= len(Sensitivity.MarkerStyleList)
        return Sensitivity.MarkerStyleList[style]

    def one_at_a_time(self, *args, **kwargs):
        """Perform One-At-a-Time (OAT) sensitivity analysis.

        This method performs OAT sensitivity analysis by varying each parameter one at a time
        while keeping others constant. For each parameter, it generates a range of values
        between the lower and upper bounds, evaluates the model function for each value,
        and stores the results.

        The results are stored in the `sen` attribute, which is a dictionary with parameter
        names as keys. Each value is a list containing:
        1. Relative parameter values (ratio to original value)
        2. Corresponding metric values from the model function
        3. Actual parameter values used
        4. Additional calculated values (if return_values=2)

        Args:
            *args: Variable length argument list passed to the model function.
            **kwargs: Arbitrary keyword arguments passed to the model function.

        Raises:
            ValueError: If the function returns more than one value when return_values=1,
                or doesn't return the expected format when return_values=2.

        Examples:
            - Import necessary libraries:
                ```python
                >>> import pandas as pd
                >>> import numpy as np
                >>> from statista.sensitivity import Sensitivity

                ```
            - Define a model function:
                ```python
                >>> def model_function(params, multiplier=1):
                ...     return multiplier * (params[0]**2 + params[1])

                ```
            - Create parameter DataFrame:
                ```python
                >>> parameters = pd.DataFrame({'value': [2.0, 3.0]}, index=['param1', 'param2'])

                ```
            - Define parameter bounds:
                ```python
                >>> lower_bounds = [0.5, 1.0]
                >>> upper_bounds = [4.0, 5.0]

                ```
            - Create sensitivity analysis object:
                ```python
                >>> sensitivity = Sensitivity(parameters, lower_bounds, upper_bounds, model_function)

                ```
            - Perform OAT sensitivity analysis with additional argument:
                ```python
                >>> sensitivity.one_at_a_time(multiplier=2)
                0-param1 -0
                6.5
                0-param1 -1
                9.781
                0-param1 -2
                14.0
                0-param1 -3
                16.125
                0-param1 -4
                ...
                1-param2 -2
                14.0
                1-param2 -3
                14.0
                1-param2 -4
                16.0
                1-param2 -5
                18.0

                ```
            - Access results for the first parameter:
                ```python
                >>> param_name = parameters.index[0]
                >>> relative_values = sensitivity.sen[param_name][0]
                >>> metric_values = sensitivity.sen[param_name][1]
                >>> actual_values = sensitivity.sen[param_name][2]

                ```
            - Print a sample result:
                ```python
                >>> print(f"When {param_name} = {actual_values[0]}, metric = {metric_values[0]}")
                When param1 = 0.5, metric = 6.5

                ```
        """
        self.sen = {}

        for i in range(self.num_parameters):
            k = self.positions[i]
            if self.return_values == 1:
                self.sen[self.parameter.index[k]] = [[], [], []]
            else:
                self.sen[self.parameter.index[k]] = [[], [], [], []]
            # generate 5 random values between the high and low parameter bounds
            rand_value = np.linspace(
                self.lower_bound[k], self.upper_bound[k], self.NoValues
            )
            # add the value of the calibrated parameter and sort the values
            rand_value = np.sort(np.append(rand_value, self.parameter["value"][k]))
            # store the relative values of the parameters in the first list in the dict
            self.sen[self.parameter.index[k]][0] = [
                (h / self.parameter["value"][k]) for h in rand_value
            ]

            random_param = self.parameter["value"].tolist()
            for j in range(len(rand_value)):
                random_param[k] = rand_value[j]
                # args = list(args)
                # args.insert(Position, random_param)
                if self.return_values == 1:
                    metric = self.function(random_param, *args, **kwargs)
                else:
                    metric, calculated_values = self.function(
                        random_param, *args, **kwargs
                    )
                    self.sen[self.parameter.index[k]][3].append(calculated_values)
                try:
                    # store the metric value in the second list in the dict
                    self.sen[self.parameter.index[k]][1].append(round(metric, 3))
                except TypeError:
                    message = """the Given function returns more than one value,
                    the function should return only one value for return_values=1, or
                    two values for return_values=2.
                    """
                    raise ValueError(message)
                # store the real values of the parameter in the third list in the dict
                self.sen[self.parameter.index[k]][2].append(round(rand_value[j], 4))
                print(str(k) + "-" + self.parameter.index[k] + " -" + str(j))
                print(round(metric, 3))

    def sobol(
        self,
        real_values: bool = False,
        title: str = "",
        xlabel: str = "xlabel",
        ylabel: str = "Metric values",
        labelfontsize=12,
        plotting_from="",
        plotting_to="",
        title2: str = "",
        xlabel2: str = "xlabel2",
        ylabel2: str = "ylabel2",
        spaces=None,
    ):
        """Plot sensitivity analysis results using Sobol-style visualization.

        This method creates plots to visualize the results of sensitivity analysis.
        It can generate either a single plot (when return_values=1) or two plots
        (when return_values=2) to show both the metric values and additional calculated values.

        Args:
            real_values (bool, optional):
                If True, plots actual parameter values on x-axis instead of relative values. Works best when
                analyzing a single parameter since parameter ranges may differ. Defaults to False.
            title (str, optional):
                Title for the main plot. Defaults to "".
            xlabel (str, optional):
                X-axis label for the main plot. Defaults to "xlabel".
            ylabel (str, optional):
                Y-axis label for the main plot. Defaults to "Metric values".
            labelfontsize (int, optional):
                Font size for axis labels. Defaults to 12.
            plotting_from (str or int, optional):
                Starting index for plotting calculated values in the second plot. Defaults to "" (start from beginning).
            plotting_to (str or int, optional): Ending index for plotting calculated values
                in the second plot. Defaults to "" (plot until end).
            title2 (str, optional): Title for the second plot (when return_values=2).
                Defaults to "".
            xlabel2 (str, optional):
                X-axis label for the second plot. Defaults to "xlabel2".
            ylabel2 (str, optional):
                Y-axis label for the second plot. Defaults to "ylabel2".
            spaces (List[float], optional):
                Spacing parameters for subplot adjustment [left, bottom, right, top, wspace, hspace]. Defaults to None.

        Returns:
            tuple: When return_values=1, returns (fig, ax) where fig is the matplotlib figure
                and ax is the axis. When return_values=2, returns (fig, (ax1, ax2)) where
                ax1 is the main plot axis and ax2 is the calculated values plot axis.

        Raises:
            ValueError:
                If attempting to plot calculated values when return_values is not 2.

        Examples:
            - Import necessary libraries:
                ```python
                >>> import pandas as pd
                >>> import numpy as np
                >>> from statista.sensitivity import Sensitivity
                >>> import matplotlib.pyplot as plt

                ```
            - Define a model function:
                ```python
                >>> def model_function(params):
                ...     return params[0]**2 + params[1]

                ```
            - Create parameter DataFrame:
                ```python
                >>> parameters = pd.DataFrame({'value': [2.0, 3.0]}, index=['param1', 'param2'])

                ```
            - Define parameter bounds:
                ```python
                >>> lower_bounds = [0.5, 1.0]
                >>> upper_bounds = [4.0, 5.0]

                ```
            - Create sensitivity analysis object:
                ```python
                >>> sensitivity = Sensitivity(parameters, lower_bounds, upper_bounds, model_function)

                ```
            - Perform OAT sensitivity analysis:
                ```python
                >>> sensitivity.one_at_a_time()
                0-param1 -0
                3.25
                0-param1 -1
                4.891
                0-param1 -2
                7.0
                0-param1 -3
                ...
                1-param2 -2
                7.0
                1-param2 -3
                7.0
                1-param2 -4
                8.0
                1-param2 -5
                9.0

                ```
            - Plot results with relative parameter values:
                ```python
                >>> fig, ax = sensitivity.sobol(
                ...     title="Parameter Sensitivity Analysis",
                ...     xlabel="Relative Parameter Value",
                ...     ylabel="Model Output"
                ... )
                >>> plt.show()
                ```
                ![one-at-a-time](./../_images/sensitivity/one-at-a-time.png)

            - Plot results with actual parameter values:
                ```python
                >>> fig2, ax2 = sensitivity.sobol(
                ...     real_values=True,
                ...     title="Parameter Sensitivity Analysis",
                ...     xlabel="Parameter Value",
                ...     ylabel="Model Output"
                ... )

                ```
                ![one-at-a-time](./../_images/sensitivity/real_values.png)
        """
        if self.return_values == 1:
            fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(8, 6))

            for i in range(self.num_parameters):
                k = self.positions[i]
                if real_values:
                    ax.plot(
                        self.sen[self.parameter.index[k]][2],
                        self.sen[self.parameter.index[k]][1],
                        Sensitivity.marker_style(k),
                        linewidth=3,
                        markersize=10,
                        label=self.parameter.index[k],
                    )
                else:
                    ax.plot(
                        self.sen[self.parameter.index[k]][0],
                        self.sen[self.parameter.index[k]][1],
                        Sensitivity.marker_style(k),
                        linewidth=3,
                        markersize=10,
                        label=self.parameter.index[k],
                    )

            ax.set_title(title, fontsize=12)
            ax.set_xlabel(xlabel, fontsize=12)
            ax.set_ylabel(ylabel, fontsize=12)

            ax.tick_params(axis="both", which="major", labelsize=labelfontsize)

            ax.legend(fontsize=12)
            plt.tight_layout()
            return fig, ax
        else:  # self.return_values == 2 and CalculatedValues
            try:
                fig, (ax1, ax2) = plt.subplots(ncols=1, nrows=2, figsize=(8, 6))

                for i in range(self.num_parameters):
                    # for i in range(len(self.sen[self.parameter.index[0]][0])):
                    k = self.positions[i]
                    if real_values:
                        ax1.plot(
                            self.sen[self.parameter.index[k]][2],
                            self.sen[self.parameter.index[k]][1],
                            Sensitivity.marker_style(k),
                            linewidth=3,
                            markersize=10,
                            label=self.parameter.index[k],
                        )
                    else:
                        ax1.plot(
                            self.sen[self.parameter.index[k]][0],
                            self.sen[self.parameter.index[k]][1],
                            Sensitivity.marker_style(k),
                            linewidth=3,
                            markersize=10,
                            label=self.parameter.index[k],
                        )

                ax1.set_title(title, fontsize=12)
                ax1.set_xlabel(xlabel, fontsize=12)
                ax1.set_ylabel(ylabel, fontsize=12)
                ax1.tick_params(axis="both", which="major", labelsize=labelfontsize)

                ax1.legend(fontsize=12)

                for i in range(self.num_parameters):
                    k = self.positions[i]
                    # for j in range(self.n_values):
                    for j in range(len(self.sen[self.parameter.index[k]][0])):
                        if plotting_from == "":
                            plotting_from = 0
                        if plotting_to == "":
                            plotting_to = len(
                                self.sen[self.parameter.index[k]][3][j].values
                            )

                        ax2.plot(
                            self.sen[self.parameter.index[k]][3][j].values[
                                plotting_from:plotting_to
                            ],
                            label=self.sen[self.parameter.index[k]][2][j],
                        )

                # ax2.legend(fontsize=12)
                box = ax2.get_position()
                ax2.set_position([box.x0, box.y0, box.width * 0.8, box.height])
                ax2.legend(loc=6, fancybox=True, bbox_to_anchor=(1.015, 0.5))

                ax2.set_title(title2, fontsize=12)
                ax2.set_xlabel(xlabel2, fontsize=12)
                ax2.set_ylabel(ylabel2, fontsize=12)

                plt.subplots_adjust(
                    left=spaces[0],
                    bottom=spaces[1],
                    right=spaces[2],
                    top=spaces[3],
                    wspace=spaces[4],
                    hspace=spaces[5],
                )

            except ValueError:
                assert ValueError(
                    "To plot calculated values, you should choose return_values==2 in the sensitivity object"
                )

            plt.tight_layout()
            plt.show()
            return fig, (ax1, ax2)



""" Tests for the eva module. """

import matplotlib
import pandas as pd

matplotlib.use("Agg")
from pandas import DataFrame
import shutil
from pathlib import Path
import pytest
from statista.eva import ams_analysis


@pytest.mark.slow
def test_eva(
    ams_gauges: DataFrame,
    gauges_statistical_properties: DataFrame,
    gauges_distribution_properties: DataFrame,
):
    path = Path("tests/data/gauges/figures")
    if path.exists():
        try:
            path.rmdir()
        except PermissionError:
            print("PermissionError: files were not deleted")

    method = "lmoments"
    save_to = "tests/data/gauges"
    statistical_properties, distribution_properties = ams_analysis(
        time_series_df=ams_gauges,
        ams=True,
        save_plots=True,
        save_to=save_to,
        filter_out=-9,
        method=method,
        alpha=0.05,
    )
    statistical_properties.drop(columns=["nyr"], inplace=True)
    gauges_statistical_properties.drop(columns=["nyr"], inplace=True)
    pd.testing.assert_frame_equal(
        statistical_properties, gauges_statistical_properties, rtol=1e-4
    )
    pd.testing.assert_frame_equal(
        distribution_properties, gauges_distribution_properties, rtol=1e-4
    )
    # try:
    shutil.rmtree(path)



"""Confidence Interval Tests"""

from typing import Dict
import numpy as np
from statista.confidence_interval import ConfidenceInterval
from statista.distributions import GEV


def test_boot_strap(
    time_series1: list,
    ci_cdf: np.ndarray,
    ci_param: Dict[str, float],
):
    """
    - The test function for the boot strap method in the ConfidenceInterval class.
    - the test can not compare the generated LB and UB as they are randomly generated.
    """
    ci = ConfidenceInterval.boot_strap(
        time_series1,
        state_function=GEV.ci_func,
        gevfit=ci_param,
        n_samples=len(time_series1),
        F=ci_cdf,
        method="lmoments",
    )
    lb = ci["lb"]
    ub = ci["ub"]
    assert isinstance(lb, np.ndarray)
    assert isinstance(ub, np.ndarray)
    assert lb.shape == ub.shape == (len(time_series1),)



""""Statistical tools"""

from typing import List, Union
import numpy as np


class Tools:
    """Collection of statistical and data transformation utilities.

    This class provides static methods for various data transformations and 
    manipulations commonly used in statistical analysis, including normalization,
    standardization, rescaling, and logarithmic transformations.

    All methods are implemented as static methods, so they can be called directly
    without instantiating the class.

    Examples:
        - Import the Tools class:
            ```python
            >>> import numpy as np
            >>> from statista.tools import Tools
            ```
        -  Normalize an array to [0, 1] range
            ```python
            >>> data = [10, 20, 30, 40, 50]
            >>> normalized = Tools.normalize(data)
            >>> print(normalized)
            [0.   0.25 0.5  0.75 1.  ]

            ```
        - Standardize an array (mean=0, std=1):
            ```python
            >>> standardized = Tools.standardize(data)
            >>> print(f"Mean: {np.mean(standardized):.4f}, Std: {np.std(standardized):.4f}")
            Mean: 0.0000, Std: 1.0000
            ```
    """

    def __init__(self):
        pass

    @staticmethod
    def normalize(x: Union[List[float], np.ndarray]) -> np.ndarray:
        """Normalize values to the range [0, 1].

        Scales all values in the input array to the range [0, 1] using min-max normalization.
        The formula used is: (x - min(x)) / (max(x) - min(x))

        Args:
            x: Input array or list of values to normalize.

        Returns:
            np.ndarray: Array of normalized values in the range [0, 1].

        Raises:
            ValueError: If all values in the input are identical (max = min),
                which would cause division by zero.

        Examples:
            - Normalize a list of values:
                ```python
                >>> from statista.tools import Tools
                >>> data = [10, 20, 30, 40, 50]
                >>> normalized = Tools.normalize(data)
                >>> print(normalized)
                [0.   0.25 0.5  0.75 1.  ]

                ```
            - Normalize a numpy array:
                ```python
                >>> import numpy as np
                >>> data = np.array([5, 15, 25, 35, 45])
                >>> normalized = Tools.normalize(data)
                >>> print(normalized)
                [0.   0.25 0.5  0.75 1.  ]

                ```
            - Edge case: single value:
                ```python
                >>> data = [42]
                >>> normalized = Tools.normalize(data)
                >>> print(normalized)
                [0.]

                ```

        See Also:
            - Tools.standardize: For standardizing values to mean=0, std=1
            - Tools.rescale: For rescaling values to a custom range
        """
        x = np.array(x)
        data_max = max(x)
        data_min = min(x)
        return (x - data_min) / (data_max - data_min)

    @staticmethod
    def standardize(x: Union[List[float], np.ndarray]) -> np.ndarray:
        """Standardize values to have mean=0 and standard deviation=1.

        Transforms the input array so that it has a mean of 0 and a standard deviation of 1.
        This is also known as z-score normalization or standard score.
        The formula used is: (x - mean(x)) / std(x)

        Args:
            x: Input array or list of values to standardize.

        Returns:
            np.ndarray: Array of standardized values with mean=0 and std=1.

        Raises:
            ValueError: If the standard deviation of the input is zero,
                which would cause division by zero.

        Examples:
            - Standardize a list of values:
                ```python
                >>> from statista.tools import Tools
                >>> import numpy as np
                >>> data = [10, 20, 30, 40, 50]
                >>> standardized = Tools.standardize(data)
                >>> print(f"Mean: {np.mean(standardized):.4f}, Std: {np.std(standardized):.4f}")
                Mean: 0.0000, Std: 1.0000

                ```
            - Verify the transformation:
                ```python
                >>> print(standardized)
                [-1.41421356 -0.70710678  0.          0.70710678  1.41421356]

                ```
            - Standardize values that already have mean=0:
                ```python
                >>> data = [-2, -1, 0, 1, 2]
                >>> standardized = Tools.standardize(data)
                >>> print(standardized)
                [-1.41421356 -0.70710678  0.          0.70710678  1.41421356]

                ```

        Notes:
            Standardization is particularly useful for algorithms that assume
            the data is centered around zero with unit variance, such as many
            machine learning algorithms.

        See Also:
            - Tools.normalize: For scaling values to the range [0, 1]
            - Tools.rescale: For rescaling values to a custom range
        """
        x = np.array(x)

        mean = np.mean(x)
        std = np.std(x)
        s = (x - mean) / std
        return s

    @staticmethod
    def rescale(
        old_value: float, 
        old_min: float, 
        old_max: float, 
        new_min: float, 
        new_max: float
    ) -> float:
        """Rescale a value from one range to another.

        Linearly transforms a value from its original range [old_min, old_max] 
        to a new range [new_min, new_max]. This is useful for mapping values 
        between different scales while preserving their relative positions.

        The formula used is:
        new_value = (((old_value - old_min) * (new_max - new_min)) / (old_max - old_min)) + new_min

        Args:
            old_value: The value to rescale.
            old_min: The minimum value of the original range.
            old_max: The maximum value of the original range.
            new_min: The minimum value of the target range.
            new_max: The maximum value of the target range.

        Returns:
            float: The rescaled value in the new range.

        Raises:
            ZeroDivisionError: If old_max equals old_min, causing division by zero.

        Examples:
            - Rescale a value from [0, 100] to [0, 1]:
                ```python
                >>> from statista.tools import Tools
                >>> value = 75
                >>> rescaled = Tools.rescale(value, 0, 100, 0, 1)
                >>> print(rescaled)
                0.75

                ```
            - Rescale a value from [0, 1] to [-1, 1]:
                ```python
                >>> value = 0.3
                >>> rescaled = Tools.rescale(value, 0, 1, -1, 1)
                >>> print(rescaled)
                -0.4

                ```
            - Rescale a temperature from Celsius to Fahrenheit:
                ```python
                >>> celsius = 25
                >>> fahrenheit = Tools.rescale(celsius, 0, 100, 32, 212)
                >>> print(f"{celsius}°C = {fahrenheit}°F")
                25°C = 77.0°F
                ```

        See Also:
            - Tools.normalize: For scaling values to the range [0, 1]
            - Tools.log_rescale: For logarithmic rescaling
        """
        old_range = old_max - old_min
        new_range = new_max - new_min
        new_value = (((old_value - old_min) * new_range) / old_range) + new_min

        return new_value

    @staticmethod
    def log_rescale(
        x: float, 
        min_old: float, 
        max_old: float, 
        min_new: float, 
        max_new: float
    ) -> int:
        """Rescale a value using logarithmic transformation.

        Transforms a value from its original range to a new range using logarithmic scaling.
        This is useful when dealing with data that spans multiple orders of magnitude,
        as it compresses large values and expands small values.

        The method first converts the value and boundaries to logarithmic space,
        then performs a linear rescaling in that space, and finally rounds to an integer.

        Args:
            x: The value to rescale.
            min_old: The minimum value of the original range.
            max_old: The maximum value of the original range.
            min_new: The minimum value of the target range.
            max_new: The maximum value of the target range.

        Returns:
            int: The logarithmically rescaled value as an integer in the new range.

        Raises:
            ValueError: If max_old is not greater than min_old.
            ValueError: If x is negative (logarithm undefined).

        Examples:
            - Rescale a value from [1, 1000] to [1, 10]:
                ```python
                >>> from statista.tools import Tools
                >>> value = 100
                >>> rescaled = Tools.log_rescale(value, 1, 1000, 1, 10)
                >>> print(rescaled)
                7

                ```
            - Rescale a small value:
                ```python
                >>> value = 10
                >>> rescaled = Tools.log_rescale(value, 1, 1000, 1, 10)
                >>> print(rescaled)
                4

                ```
            - Handle zero values (special case):
                ```python
                >>> value = 0
                >>> rescaled = Tools.log_rescale(value, 0, 1000, 0, 10)
                >>> print(rescaled)
                0

                ```

        Notes:
            - For x = 0, the function uses a special case handling by setting the log value to -7.
            - For min_old = 0, the function also uses -7 as the logarithmic value.
            - The base of the logarithm is e (natural logarithm).

        See Also:
            - Tools.inv_log_rescale: For inverse logarithmic rescaling
            - Tools.rescale: For linear rescaling
        """
        # get the boundaries of the logarithmic scale
        if np.isclose(min_old, 0.0):
            min_old_log = -7
        else:
            min_old_log = np.log(min_old)

        max_old_log = np.log(max_old)

        if x == 0:
            x_log = -7
        else:
            x_log = np.log(x)

        y = int(
            np.round(Tools.rescale(x_log, min_old_log, max_old_log, min_new, max_new))
        )

        return y

    @staticmethod
    def inv_log_rescale(
        x: float, 
        min_old: float, 
        max_old: float, 
        min_new: float, 
        max_new: float, 
        base: float = np.e
    ) -> int:
        """Rescale a value using inverse logarithmic transformation.

        Performs the inverse operation of log_rescale. Instead of taking logarithms,
        this method raises the base to the power of the input values before rescaling.
        This is useful when you need to expand the scale of values that were previously
        compressed using a logarithmic transformation.

        The method first converts the value and boundaries to exponential space using
        the specified base, then performs a linear rescaling in that space, and finally
        rounds to an integer.

        Args:
            x: The value to rescale.
            min_old: The minimum value of the original range.
            max_old: The maximum value of the original range.
            min_new: The minimum value of the target range.
            max_new: The maximum value of the target range.
            base: The base to use for the exponential transformation. Defaults to e (natural exponential).

        Returns:
            int: The inverse logarithmically rescaled value as an integer in the new range.

        Raises:
            ValueError: If max_old is not greater than min_old.
            OverflowError: If the exponential values are too large to handle.

        Examples:
            - Rescale a value from [1, 3] to [1, 1000] using base e:
                ```python
                >>> from statista.tools import Tools
                >>> value = 2
                >>> rescaled = Tools.inv_log_rescale(value, 1, 3, 1, 1000)
                >>> print(rescaled)
                148

                ```

            - Using a different base (base 10):
                ```python
                >>> import numpy as np
                >>> value = 1
                >>> rescaled = Tools.inv_log_rescale(value, 0, 2, 1, 100, base=10)
                >>> print(rescaled)
                10

                ```
            - Verify inverse relationship with log_rescale:
                ```python
                >>> original = 500
                ```
            - First log_rescale from [1, 1000] to [0, 3]:
                ```python
                >>> log_scaled = Tools.log_rescale(original, 1, 1000, 0, 3)

                ```
            - Then inv_log_rescale back from [0, 3] to [1, 1000]:
                ```python
                >>> back_to_original = Tools.inv_log_rescale(log_scaled, 0, 3, 1, 1000)
                >>> print(f"Original: {original}, After round-trip: {back_to_original}")
                Original: 500, After round-trip: 403

                ```

        Notes:
            Due to rounding and the discrete nature of the transformation,
            the round-trip conversion (log_rescale followed by inv_log_rescale)
            may not exactly reproduce the original value.

        See Also:
            - Tools.log_rescale: For logarithmic rescaling
            - Tools.rescale: For linear rescaling
        """
        # get the boundaries of the logarithmic scale

        min_old_power = np.power(base, min_old)
        max_old_power = np.power(base, max_old)
        x_power = np.power(base, x)

        y = int(
            np.round(
                Tools.rescale(x_power, min_old_power, max_old_power, min_new, max_new)
            )
        )
        return y

    @staticmethod
    def round(number: float, precision: float) -> float:
        """Round a number to a specified precision.

        Rounds a number to the nearest multiple of the specified precision.
        This is different from Python's built-in round function, which rounds
        to a specified number of decimal places.

        Args:
            number: The number to be rounded.
            precision: The precision to round to. For example, if precision is 0.5,
                the number will be rounded to the nearest 0.5.

        Returns:
            float: The rounded number.

        Examples:
            - Round to the nearest 0.5
            ```python
            >>> from statista.tools import Tools
            >>> value = 3.7
            >>> rounded = Tools.round(value, 0.5)
            >>> print(rounded)
            3.5

            ```

            - Round to the nearest 5:
                ```python
                >>> value = 23
                >>> rounded = Tools.round(value, 5)
                >>> print(rounded)
                25

                ```

            - Round to the nearest 0.1:
                ```python
                >>> value = 7.84
                >>> rounded = Tools.round(value, 0.1)
                >>> print(rounded)
                7.8

                ```

        Notes:
            The formula used is: round(number / precision) * precision

            This method is useful for rounding to specific increments rather than
            decimal places. For example, rounding to the nearest 0.25, 0.5, or 5.
        """
        return round(number / precision) * precision



try:
    from importlib.metadata import PackageNotFoundError, version  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import PackageNotFoundError, version  # type: ignore


try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"

# documentation format
__author__ = "Mostafa Farrag"
__email__ = "moah.farag@gmail.com"
__docformat__ = "restructuredtext"

__metadata__ = {
    "author": __author__,
    "email": __email__,
    "version": __version__,
    "docformat": __docformat__,
}


__doc__ = """
statista - statistics package
"""
